with Ada.Text_IO, Strings, Rationals, Flts, Globals, Param, Reps;
use Ada.Text_IO, Strings, Rationals, Flts, Globals, Param, Reps;
with Matrices, Vectors, ScalVectors, Fouriers2, Fouriers2.BBeam;

pragma Elaborate_All (Ada.Text_IO,Strings,Rationals,Flts,Globals,Param,Reps);
pragma Elaborate_All (Matrices,Vectors,ScalVectors,Fouriers2,Fouriers2.BBeam);

procedure Approx_BFix(V: in Pars) is
  
  subtype Scalar is Rep;

  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Scalar => Scalar, Component => Scalar, SM => SM);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2 is new Fouriers2 (Deg1 => V.Deg1, Deg2 => V.Deg2, Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2_Beam is new SFouriers2.BBeam (PDeg1 => V.PDeg1, PDeg2 => V.PDeg2, SSV => SSV);
  use SM, SFouriers2, SFouriers2_Beam, SFouriers2_Beam.BContr;
  
  BB:     constant Positive := INum(V.Kappa);
  PostFix: constant String := "fix_" & Strng(BB);
  PostMat: constant String := "mat_" & Strng(BB) & "." & Strng0(VDim,3);
  
  procedure Rough_Fix(F: in out Fourier2) is
  begin
    SetZero(F);
    F.T.R := (V.Rho1,V.Rho2);
    F.T.P := (0,1);
    F.C(1,1) := Scal(-1.12);
  end Rough_Fix;
  
  procedure Read_Fix(F: in out Fourier2) is
  begin
    Read(Output_Dir & "numb" & PostFix,F);
  exception
    when Name_Error => Rough_Fix(F);
  end Read_Fix;
  
  procedure Write_Fix(F: in Fourier2) is
  begin
    Write(Output_Dir & "numb" & PostFix,F);
  end Write_Fix;
  
  procedure Read_Mat(A: in out PMat) is
  begin
    Read(Output_Dir & "b" & PostMat,A);
  exception
    when Name_Error => SetZero(A);
  end Read_Mat;
  
  procedure Write_Mat(A: in PMat) is
  begin
    Write(Output_Dir & "b" & PostMat,A);
  end Write_Mat;
  
  Eps:  constant Radius := 1.0e-17;
  Eps0: constant Radius := 1.0e-12;
  
  Err,Nf: Radius;
  A: PMat;
  F,G: Fourier2;
  
begin
  
  Iter := 200;
  Read_Mat(A);
  Read_Fix(F);

  if Verbosity>1 then
    MapB(BB,F,G);
    Sub(F,G);
    Err := Sup(Norm(G));
    Nf := Sup(Norm(F));
    Show1("old Err =",Err,False);
    Show1("   norm =",Nf);
  end if;
  if IsZero(A) then
    ApproxFix(BB,F,A,Eps);
  else
    IterContr(BB,F,A,Eps);
  end if;
  MapB(BB,F,G);
  Sub(F,G);
  Err := Sup(Norm(G));
  if Verbosity>0 then
    Show1("new Err =",Err,False);
    Show1("   norm =",Nf);
  end if;
  if Err<Eps0 then
    Write_Mat(A);
    Write_Fix(F);
  end if;
  
end Approx_BFix;
