with Strings, Rationals, Flts, Globals, Param, Flts.Std.Balls;
use Strings, Rationals, Flts, Globals, Param,  Flts.Std.Balls;
with Matrices, Vectors, ScalVectors, Fouriers2, Fouriers2.BBeam;
with Reps, Approx_BFix;

pragma Elaborate_All (Strings,Rationals,Flts,Globals,Param,Flts.Std.Balls);
pragma Elaborate_All (Matrices,Vectors,ScalVectors,Fouriers2,Fouriers2.BBeam);
pragma Elaborate_All (Reps,Approx_BFix);

procedure Check_BFix(V: in Pars) is
  
  subtype Scalar is Ball;
  
  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Scalar => Scalar, Component => Scalar, SM => SM);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2 is new Fouriers2 (Deg1 => V.Deg1, Deg2 => V.Deg2, Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2_Beam is new SFouriers2.BBeam (PDeg1 => V.PDeg1, PDeg2 => V.PDeg2, SSV => SSV);
  use  SM, SFouriers2, SFouriers2_Beam, SFouriers2_Beam.BContr;
  
  KMax:    constant Radius   := Flt(3)/Flt(4);
  BB:      constant Positive := INum(V.Kappa);
  PostFix: constant String   := "fix_" & Strng(BB);
  PostMat: constant String   := "mat_" & Strng(BB) & "." & Strng0(VDim,3);
  
  procedure Make_Approx(F: in out Fourier2; A: in out PMat) is
    package Balls renames Flts.Std.Balls;
  begin
    Reps.Proper_Rounding;
    Approx_BFix(V);
    NumRead(Output_Dir & "numb" & PostFix,F);
    F.T.R := (V.Rho1,V.Rho2);
    NumRead(Output_Dir & "b" & PostMat,A);
    Balls.Proper_Rounding;
  end Make_Approx;
  
  OK: Boolean;
  C11: Scalar;
  A: PMat;
  F: Fourier2;

begin
  SetEDegs(V.EDeg1,V.EDeg2);
  Show_Degs;
  Show2("Rho",V.Rho1,V.Rho2);
  
  Make_Approx(F,A);
  CheckFix(BB,F,A,KMax,OK);
  if OK then
    Write(Output_Dir & "b" & PostFix,F);
  end if;
  
  CheckType11(F,C11,OK);
  if not (OK and then MaxNorm(C11)>One) then
    Show0("branch check failed");
  end if;
  
end Check_BFix;
