with Strings, Rationals, Flts, Globals, Param, Flts.Std.Balls;
use Strings, Rationals, Flts, Globals, Param,  Flts.Std.Balls;
with Matrices, Vectors, ScalVectors, Fouriers2, Fouriers2.NBeam;
with Reps, Approx_KFix;

pragma Elaborate_All (Strings,Rationals,Flts,Globals,Param,Flts.Std.Balls);
pragma Elaborate_All (Matrices,Vectors,ScalVectors,Fouriers2,Fouriers2.NBeam);
pragma Elaborate_All (Reps,Approx_KFix);

procedure Check_NFix(HPos: in Boolean; V: in Pars; OK: out Boolean) is
  
  subtype Scalar is Ball;
  
  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Scalar => Scalar, Component => Scalar, SM => SM);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2 is new Fouriers2 (Deg1 => V.Deg1, Deg2 => V.Deg2, Scalar => Scalar, SM => SM, SV => SV);
  package SFouriers2_Beam is new SFouriers2.NBeam (HPos => HPos, PDeg1 => V.PDeg1, PDeg2 => V.PDeg2, SSV => SSV);
  use  SM, SFouriers2, SFouriers2_Beam, SFouriers2_Beam.NContr;
  
  Kappa: Rational renames V.Kappa;
  
  KMax:    constant Radius := Flt(3)/Flt(4);
  R0:     constant RadPair := (Zero,Zero);
  N:    constant Character := Choose(HPos,'N','n');
  PostFix: constant String := "fix_" & Strng(Kappa);
  PostMat: constant String := "mat_" & Strng(Kappa) & "." & Strng0(VDim,3);
  
  procedure Make_Approx(F: in out Fourier2; A: in out PMat) is
    package Balls renames Flts.Std.Balls;
    K: constant Character := Choose(HPos,'K','k');
  begin
    Reps.Proper_Rounding;
    Approx_KFix(HPos,V);
    NumRead(Output_Dir & "num" & K & PostFix,F);
    F.T.R := (V.Rho1,V.Rho2);
    NumRead(Output_Dir & K & PostMat,A);
    Balls.Proper_Rounding;
  end Make_Approx;
  
  function CheckCoeff(F: in Fourier2) return Boolean is
    S: constant Scalar := Two*abs(Scal(1)-Inv(Beta(Kappa)));
    B: Boolean;
    C11: Scalar;
  begin
    CheckType11(F,C11,B);
    return B and then (Sqr(C11)>S);
  end CheckCoeff;
  
  A: PMat;
  F: Fourier2;

begin
  SetEDegs(V.EDeg1,V.EDeg2);
  Show_Degs;
  Show1("kappa =",Kappa);
  Show1("beta",Beta(Kappa));
  Show2("Rho",V.Rho1,V.Rho2);
  
  Make_Approx(F,A);
  CheckFix(Kappa,F,A,KMax,OK);
  if OK then
    Show1("norm0 =",Norm(R0,F));
    Write(Output_Dir & N & PostFix,F);
    OK := CheckCoeff(F);
    if not OK then Show0("branch check failed"); end if;
  end if;
  
end Check_NFix;
