with FPU_Rounding;

  pragma Elaborate_All (FPU_Rounding);

package body Flts.Std.Intervals is
  
  function PowerDown(I: Integer; X: Rep) return Rep is
  begin
    if X<RepZero and then (I mod 2)=1 then
      return -PowerUp(I,-X);
    end if;
    if X=RepZero then
      if I <= 0 then raise Undefined; end if;
      return RepZero;
    end if;
    declare
      J: Integer := abs(I);
    begin
      if J>1 then
        if I=2 then return -((-X)*X); end if;
        declare
          P,Y: Rep; -- negative
        begin
          if I>0 then
            P := -abs(X);
          else
            P := RepNegOne/abs(X);
            if J=2 then return -((-P)*P); end if;
          end if;
          if (J rem 2)=0 then
            Y := RepNegOne;
          else
            Y := P;
          end if;
          J := J/2;
          while J>0 loop
            P := (-P)*P;
            if (J rem 2) /= 0 then
              Y := (-P)*Y;
            end if;
            J := J/2;
          end loop;
          return -Y;
        end;
      elsif I=0 then
        return RepOne;
      elsif I=1 then
        return X;
      else                    --- I = -1
        return -(RepNegOne/X);
      end if;
    end;
  end PowerDown;
    
  function PowerUp(I: Integer; X: Rep) return Rep is
  begin
    if X<RepZero and then (I mod 2)=1 then
      return -PowerDown(I,-X);
    end if;
    if X=RepZero then
      if I <= 0 then raise Undefined; end if;
      return RepZero;
    end if;
    declare
      J: Integer := abs(I);
    begin
      if J>1 then
        if I=2 then return X*X; end if;
        declare
          P,Y: Rep; -- positive
        begin
          if I>0 then
            P := abs(X);
          else
            P := RepOne/abs(X);
            if J=2 then return P*P; end if;
          end if;
          if (J rem 2)=0 then
            Y := RepOne;
          else
            Y := P;
          end if;
          J := J/2;
          while J>0 loop
            P := P*P;
            if (J rem 2) /= 0 then
              Y := P*Y;
            end if;
            J := J/2;
          end loop;
          return Y;
        end;
      elsif I=0 then
        return RepOne;
      elsif I=1 then
        return X;
      else               --- I = -1
        return RepOne/X;
      end if;
    end;
  end PowerUp;
  
-----------------------------------------------------  
  
  function Scal(R: Flt) return Interval is
  begin
    return (Rep(R),Rep(R));
  end Scal;
  
  function Maxabs(S: Interval) return Rep is
  begin
    if S.L >= RepZero then
      return S.U;
    elsif S.U <= RepZero then
      return -S.L;
    else
      return Rep'Max(-S.L,S.U);
    end if;
  end Maxabs;
  
  function "abs"(S: Interval) return Interval is
  begin
    if S.L >= RepZero then
      return S;
    elsif S.U <= RepZero then
      return -S;
    else
      return (RepZero,Rep'Max(-S.L,S.U));
    end if;
  end "abs";
  
  function Union(S1,S2: Interval) return Interval is
  begin
    return (Rep'Min(S1.L,S2.L),Rep'Max(S1.U,S2.U));
  end Union;
  
  function "-"(S: Interval) return Interval is
  begin
    return (-S.U,-S.L);
  end "-";
  
  function "+"(S1,S2: Interval) return Interval is
  begin
    return (-(-S1.L-S2.L),S1.U+S2.U);
  end "+";
  
  function "-"(S1,S2: Interval) return Interval is
  begin
    return (-(S2.U-S1.L),S1.U-S2.L);
  end "-";
  
  function "*"(R: Rep; S: Interval) return Interval is
  begin
    if R<RepZero then
      return (-((-R)*S.U),R*S.L);
    else
      return (-((-R)*S.L),R*S.U);
    end if;
  end "*";
  
  function "*"(S1,S2: Interval) return Interval is
    P: constant Interval := S1.L*S2;
    Q: constant Interval := S1.U*S2;
  begin
    return (Rep'Min(P.L,Q.L),Rep'Max(P.U,Q.U));
  end "*";
  
  function Sqr(S: Interval) return Interval is
  begin
    if S.L >= RepZero then
      return (-((-S.L)*S.L),S.U*S.U);
    elsif S.U <= RepZero then
      return (-((-S.U)*S.U),S.L*S.L);
    else
      return (RepZero,Rep'Max(S.L*S.L,S.U*S.U));
    end if;
  end Sqr;

  function "/"(S: Interval; R: Rep) return Interval is
  begin
    if R<RepZero then
      return (-(S.U/(-R)),S.L/R);
    else
      return (-(S.L/(-R)),S.U/R);
    end if;
  end "/";
  
  function "/"(R: Rep; S: Interval) return Interval is
  begin
    if R<RepZero then
      return (-R)/(-S);
    elsif S.L>RepZero then
      return (-((-R)/S.U),R/S.L);
    elsif S.U<RepZero then
      return (-((-R)/S.U),R/S.L);
    else
      raise Undefined;
      return (RepZero,RepZero);
    end if;
  end "/";
  
  function "/"(S1,S2: Interval) return Interval is
    P: constant Interval := S1.L/S2;
    Q: constant Interval := S1.U/S2;
  begin
    return (Rep'Min(P.L,Q.L),Rep'Max(P.U,Q.U));
  end "/";
  
  function "**"(S: Interval; I: Integer) return Interval is
  begin
    if (I mod 2)=1 then                           -- I odd (order preserving)
      return (PowerDown(I,S.L),PowerUp(I,S.U));
    elsif S.L > RepZero then                      -- I even, S>0
      return (PowerDown(I,S.L),PowerUp(I,S.U));
    elsif S.U < RepZero then                      -- I even, S<0
      return (PowerDown(I,-S.U),PowerUp(I,-S.L));
    elsif I <= 0 then
      raise Undefined;
    end if;                                       -- I even, S contains 0
    return (RepZero,PowerUp(I,Rep'Max(-S.L,S.U)));
  end "**";
  
  function Sqrt(S: Interval) return Interval is
--- IEEE 754 Sqrt has error <= ULP/2
  begin
    if S.L > RepZero then
      return (Rep'Pred(Sqrt(S.L)),Sqrt(S.U));
    elsif S.L < RepZero then
      raise Undefined;
    end if;
    return (RepZero,Sqrt(S.U));
  end Sqrt;
  
  function PiHalf return Interval is
--- Newton
    Iter: constant Integer := LLFloat'Machine_Mantissa;
    Y: Interval := (RepOne,RepTwo);
  begin
    for I in reverse 1 .. Iter loop
      Y := (RepOne,RepOne)+LLFloat(I)*(Y/LLFloat(I+I+1));
    end loop;
    return Y;
  end PiHalf;
  
  procedure Proper_Rounding(Dummy: in Interval) is
    pragma Unreferenced(Dummy);
  begin
    FPU_Rounding.Set_Rounding_Mode((Up,Strict));
  end Proper_Rounding;
  
begin
  
  Flts.A_Numeric_Mode := FPU_Rounding.Default_Rounding_Mode'Access;
  FPU_Rounding.Set_Rounding_Mode((Up,Strict));

end Flts.Std.Intervals;
