
package body Fouriers2.BBeam is
  
  use SM,SV,SSV;
  
  function Phi1(BB: Positive; K: Natural) return Integer is
  begin
    if BB /= 3 then raise Not_Implemented; end if;
    if K  =  0 then return  0; end if;
    if K <=  2 then return  1; end if; --- 3*2^4-7^2=-1
    if K <=  3 then return 13; end if; --- 3*3^4-16^2=-13
    if K <=  7 then return 22; end if; --- 3*7^4-85^2=-22
    if K <=  8 then return 33; end if; --- 3*8^4-111^2=-33
    if K <= 88 then return 39; end if; --- 3*88^4-13413^2=39
    return                118;         --- 3*613^4-650851^2=-118
  end Phi1;
  
  function Phi2(BB: Positive; N: Natural) return Integer is
    K: Integer;
    A: LInt;  
  begin
    if N=0 then return 0; end if;
    A := LFloor(Sqrt(Scal(N)/Sqrt(Scal(BB))));
    K := Integer(A);
    A := A*A;
    A := abs(LInt(BB)*A*A-LInt(N*N));
    return IMin(Integer(A),Phi1(BB,K+1));
  end Phi2;
  
  procedure InvLinear(BB: in Positive; F: in out Fourier2) is
--- assumes Flt operations round up
    FBB: constant Flt := Flt(BB);
    K1,N1,DK,DN: Integer;
    A,B: LInt;
    C: FCoeff renames F.C;
    E: FErr renames F.E;
  begin
    for K in Freq2 loop
      A := LInt(K*K);
      A := LInt(BB)*A*A;
      for N in Freq1 loop
        if not IsZero(C(N,K)) then
          if (K mod 2)=0 or (N mod 2)=0 then raise Not_Implemented; end if;
          B := LInt(N*N);
          C(N,K) := FBB*C(N,K)/Flt(A-B);
        end if;
      end loop;
    end loop;
    if not STrunc then
      for N in 0 .. DegE1 loop
        N1 := 2*(N/2)+1;
        DN := Phi2(BB,N1);
        for K in 0 .. DegE2 loop
          if E(N,K) /= Zero then
            K1 := 2*(K/2)+1;
            DK := Phi1(BB,K1);
            E(N,K) := FBB*E(N,K)/Flt(IMax(DK,DN));
          end if;
        end loop;
      end loop;
    end if;
  end InvLinear;
  
  procedure MapB(BB: in Positive; F: in Fourier2; G: in out Fourier2) is
  begin
    Prod(F,F,G);
    Mult(F,G);
    InvLinear(BB,G);
    if HPos then Neg(G); end if;
  end MapB;

  procedure DMapB(BB: in Positive; F: in Fourier2; H: in out Fourier2) is
  begin
    Mult(F,H);
    Mult(F,H);
    InvLinear(BB,H);
    if HPos then
      Mult(-Three,H);
    else
      Mult(Three,H);
    end if;
  end DMapB;
  
end Fouriers2.BBeam;
