with Strings;
use Strings;

pragma Elaborate_All (Strings);

package body Fouriers2.KBeam is
  
  use SM,SV,SSV;
  
  procedure InvLinear(Kappa: in Rational; F: in out Fourier2) is
--- assumes Flt operations round up
    P: constant Positive := Num(Kappa);
    Q: constant Positive := Den(Kappa);
    QQ: constant Flt := Flt(Q*Q);
    K1,N1: Integer;
    A,B: LInt;
    C: FCoeff renames F.C;
    E: FErr renames F.E;
  begin
    if (P mod 2)=0 or (Q mod 2)=1 then raise Not_Implemented; end if;
    for K in Freq2 loop
      A := LInt(Q*K*K);
      A := A*A;
      for N in Freq1 loop
        if not IsZero(C(N,K)) then
          if (K mod 2)=0 or (N mod 2)=0 then raise Not_Implemented; end if;
          B := LInt(P*N);
          B := B*B;
          C(N,K) := QQ*C(N,K)/Flt(A-B);
        end if;
      end loop;
    end loop;
    if not STrunc then
      for K in 0 .. DegE2 loop
        K1 := 2*(K/2)+1;
        A := LInt(Q*K1*K1);
        for N in 0 .. DegE1 loop
          if E(N,K) /= Zero then
            N1 := 2*(N/2)+1;
            B := LInt(P*N1);
	    E(N,K) := QQ*E(N,K)/Flt(2*LInt'Max(A,B)-1);
          end if;
        end loop;
      end loop;
    end if;
  end InvLinear;
 
  procedure MapK(Kappa: in Rational; F: in Fourier2; G: in out Fourier2) is
  begin
    Prod(F,F,G);
    Mult(F,G);
    InvLinear(Kappa,G);
    if HPos then Neg(G); end if;
  end MapK;

  procedure DMapK(Kappa: in Rational; F: in Fourier2; H: in out Fourier2) is
  begin
    Mult(F,H);
    Mult(F,H);
    InvLinear(Kappa,H);
    if HPos then
      Mult(-Three,H);
    else
      Mult(Three,H);
    end if;
  end DMapK;
  
------------ numerical
  
  procedure Find_Fix(Kap: in Rational; F: in out Fourier2; A: in out PMat; OK: out Boolean) is
    Eps:  constant Radius := 1.0e-17;
    Eps0: constant Radius := 1.0e-12;
    Err,Nf: Radius;
    G: Fourier2;
  begin
    if Verbosity>1 then
      Show1("kappa = ",Kap);
      Nf := Sup(Norm(F));
      MapK(Kap,F,G);
      Sub(F,G);
      Err := Sup(Norm(G));
      Show1("old Err =",Err,False);
      Show1("   norm =",Nf);
    end if;
    if IsZero(A) then
      ApproxFix(Kap,F,A,Eps);
    else
      IterContr(Kap,F,A,Eps);
    end if;
    MapK(Kap,F,G);
    Sub(F,G);
    Err := Sup(Norm(G));
    if Verbosity>0 then
      Nf := Sup(Norm(F));
      Show1("new Err =",Err,False);
      Show1("   norm =",Nf);
    end if;
    OK := (Err<Eps0);
  end Find_Fix;
  
  function Norm0(F: Fourier2) return Radius is
  begin
    return ProjNorm0(1,1,F);
  end Norm0;
  
  function ProjNorm0(NFac,KFac: Positive; F: Fourier2) return Radius is
    DN: constant Integer := 2*NFac;
    DK: constant Integer := 2*KFac;
    N,K: Integer;
    R: Radius := Zero;
  begin
    N := NFac;
    while N <= PDeg1 loop
      K := KFac;
      while K <= PDeg2 loop
        R := R+MaxNorm(F.C(N,K));
        K := K+DK;
      end loop;
      N := N+DN;
    end loop;
    return R;
  end ProjNorm0;
  
  procedure Plot3(FileName: in String; R1,R2: in Flt; I: in Integer) is
    package Txt_IO renames Ada.Text_IO;
    F: File_Type;
  begin
    begin
      Open(F,Append_File,FileName);
    exception
      when Name_Error => Create(F,Out_File,FileName);
    end;
    Flt_IO.Put(F,R1,3,5,3);
    Txt_IO.Put(F," ");
    Flt_IO.Put(F,R2,3,5,3);
    Txt_IO.Put(F,"  " & Strng(I));
    New_Line(F);
    Close(F);
  end Plot3;
    
end Fouriers2.KBeam;
