with Strings;
use Strings;

pragma Elaborate_All (Strings);

package body Fouriers2.Wave is
   
  procedure SetZeroKernel(Kappa: in Rational; F: in out Fourier2) is
    N: constant Positive := Num(Kappa);
    D: constant Positive := Den(Kappa);
    FC: FCoeff renames F.C;
  begin
    for K in 0 .. IMin(Deg1/D,Deg2/N) loop
      SetZero(FC(K*D,K*N));
    end loop;
  end SetZeroKernel;
  
  procedure AddKernel(Kappa: in Rational; F: in Fourier2; G: in out Fourier2) is
    N: constant Positive := Num(Kappa);
    D: constant Positive := Den(Kappa);
    FC: FCoeff renames F.C;
    GC: FCoeff renames G.C;
  begin
    if not STrunc then raise Not_Implemented; end if;
    for K in 0 .. IMin(Deg1/D,Deg2/N) loop
      Add(FC(K*D,K*N),GC(K*D,K*N));
    end loop;
  end AddKernel;
  
  function Kernel(Kappa: Rational; F: Fourier2) return Fourier2 is
    G: Fourier2;
  begin
    SetZero(F.T,G);
    AddKernel(Kappa,F,G);
    return G;
  end Kernel;
  
   procedure InvLinear(Kappa: in Rational; F: in out Fourier2) is
----- assumes Kernel error-components are zero
      N: constant Positive := Num(Kappa);
      D: constant Positive := Den(Kappa);
      DD: constant Flt := Flt(D*D);
      Ni,JMin: Integer;
      AII: Scalar;
      FC: FCoeff renames F.C;
      FE: FErr renames F.E;
   begin
      for I in Freq1 loop
         NI := N*I; 
	 AII := Scal(Kappa*Kappa)*Scal(I*I);
	 for J in Freq2 loop
            if not IsZero(FC(I,J)) then
               if D*J=NI then
                  if not STrunc then raise Undefined; end if;
               else
		  FC(I,J) := FC(I,J)/(Scal(J*J)-AII);
	       end if;
	    end if;
	 end loop;
      end loop;
      if not STrunc then
         FE(0,0) := DD*FE(0,0)/Flt(IMin(N,D));
         JMin := 1;
         for I in FreqE1 loop
            NI := N*I;
            if I=1 then JMin := 0; end if;
	    for J in JMin .. DegE2 loop
	       FE(I,J) := DD*FE(I,J)/Flt(2*IMax(D*J,NI)-1);
	    end loop;
	 end loop;
      end if;
   end InvLinear;
   
   procedure MapA(Kappa: in Rational; F: in Fourier2; G: in out Fourier2) is
   begin
      Prod(F,F,G);
      Mult(F,G);
      InvLinear(Kappa,G);
      if STrunc then
        AddKernel(Kappa,F,G);
      end if;
   end MapA;

   procedure DMapA(Kappa: in Rational; F: in Fourier2; H: in out Fourier2) is
      H0: constant Fourier2 := H;
   begin
      Prod(F,H0,H);
      Mult(F,H);
      InvLinear(Kappa,H);
      Mult(Three,H);
      if STrunc then
        AddKernel(Kappa,H0,H);
      end if;
   end DMapA;
   
   procedure CheckCoeff(Kappa: in Rational; F: in Fourier2) is
     Q: constant Rational := 2*abs(1-Kappa);
     OK: Boolean;
     C11: Scalar;
   begin
     CheckType11(F,C11,OK);
     if not (OK and then Sqr(C11)>Scal(Q)) then
       Show0("branch check failed.");
     end if;
   end CheckCoeff;
     
   procedure Find_Fix(Kappa: in Rational; V1: in out Fourier2) is
      --- use only with numeric scalars
      Eps: constant Radius := 1.0e-15;
      V2: Fourier2;
   begin
      if not STrunc then
	 raise Sorry with "Fouriers2.Contr.Find_Fix: run me in numeric mode";
      end if;
      MapA(Kappa,V1,V2);
      if Verbosity >0 then
	 Show1("norm: ",Norm(V1));
	 Show1("diff: ",Norm(V2-V1));
      end if;
      ApproxFix(Kappa,V1,Eps);
      MapA(Kappa,V1,V2);
      if Verbosity >0 then
	 Show1("norm: ",Norm(V1));
	 Show1("diff: ",Norm(V2-V1));
      end if;
   end Find_Fix;

   procedure Find_Fix(Kappa: in Rational; NameIn,NameOut: in String) is
      --- use only with numeric scalars
      F: Fourier2;
   begin
      if not STrunc then
	 raise Sorry with "Fouriers2.Contr.Find_Fix: run me in numeric mode";
      end if;
      Read(NameIn,F);
      Find_Fix(Kappa,F);
      Write(NameOut,F);
   end Find_Fix;

end Fouriers2.Wave;
