with Strings, Flt_Addetc, Flt_Vectors;
use Strings, Flt_Addetc, Flt_Vectors;

pragma Elaborate_All (Strings,Flt_Addetc,Flt_Vectors);

package body Fouriers2 is
   
  use Flt_Matrices,SM,SV,SV_More,SM_More;
   
  procedure UpdateNormFacs(R: in RadPair) is
--- uses global LastPair, NormFacs1, NormFacs2, NormFacs
    begin
      if R(1) /= LastPair(1) then
        Powers(NotExp(R(1)),NormFacs1);
      elsif R(2)=LastPair(2) then
        return;
      end if;
      if R(2) /= LastPair(2) then
	 if R(2)=R(1) then
	    NormFacs2 := NormFacs1;
	 else
	    Powers(NotExp(R(2)),NormFacs2);
	 end if;
      end if;
      for M in FreqE1 loop
	 for N in FreqE2 loop
	    NormFacs(M,N) := NormFacs1(M)*NormFacs2(N);
	 end loop;
      end loop;
      LastPair := R;
   end UpdateNormFacs;
   
   procedure Powers(S: in Scalar; W: in out SV.Vector) is
      M: Integer := 1;
   begin
      W(0) := SOne;
      W(1) := S;
      for K in 2 .. W'Last loop
	 W(K) := W(K-M)*W(M);
	 if 2*M=K then M := K; end if;
      end loop;
      for K in W'First .. -1 loop
	 W(K) := Inv(W(-K));
      end loop;
   end Powers;
   
  function NotExp(X: Flt) return Scalar is
  begin
    return SOne+Scal(X);
  end NotExp;
   
   function CosSin(P: Parity; S: Scalar) return Scalar is
   begin
      if P=0 then return Cos(S); else return Sin(S); end if;
   end CosSin;

   procedure AddProd(S: in Scalar; C1: in FCoeff; C2: in out FCoeff; Tmp: in out Scalar) is
   begin
      if not IsZero(S) then
	 for I in Freq1 loop
	    for J in Freq2 loop AddProd(S,C1(I,I),C2(I,J),Tmp); end loop;
	 end loop;
      end if;
   end AddProd;
   
   procedure SubProd(S: in Scalar; C1: in FCoeff; C2: in out FCoeff; Tmp: in out Scalar) is
   begin
      if not IsZero(S) then
	 for I in Freq1 loop
	    for J in Freq2 loop SubProd(S,C1(I,I),C2(I,J),Tmp); end loop;
	 end loop;
      end if;
   end SubProd;
   
   --- basic

   function Info(Dummy: Fourier2) return Scalar_Info is
      pragma Unreferenced(Dummy);
   begin
      return FInfo;
   end Info;

   function IsPoly(P: Fourier2) return Boolean is
   begin
      return STrunc or else IsZero(P.E);
   end IsPoly;
   
   function IsSharp(P: Fourier2) return Boolean is
   begin
      return STrunc or else (IsSharp(P.C) and then IsZero(P.E));
   end IsSharp;
   
   function IsZero(P: Fourier2) return Boolean is
   begin
      return IsZero(P.C) and then (STrunc or else IsZero(P.E));
   end IsZero;

   function "="(P1,P2: Fourier2) return Boolean is
      C1: FCoeff renames P1.C;
      C2: FCoeff renames P2.C;
   begin
      if (C1 /= C2) then return False; end if;
      if STrunc then return True; end if;
      if (P1.E /= P2.E) then return False; end if;
      if IsZero(P1.E) then return True; end if;
      return (P1.T = P2.T);
   end "=";
   
  procedure SetZero(P: in out Fourier2) is
  begin
    P.T.R := Rad;
    SetZero(P.C);
    if not STrunc then SetZero(P.E); end if;
  end SetZero;
   
  procedure SetZero(T: in FType; F: in out Fourier2) is
  begin
    F.T := T;
    SetZero(F.C);
    SetZero(F.E);
  end SetZero;

  procedure SetZero(Template: in Fourier2; F: in out Fourier2) is
  begin
    F.T := Template.T;
    SetZero(F.C);
    SetZero(F.E);
  end SetZero;

  procedure SetZero(P: in ParPair; R: in RadPair; F: in out Fourier2) is
  begin
    F.T.P := P;
    F.T.R := R;
    SetZero(F.C);
    SetZero(F.E);
  end SetZero;

  procedure Copy(P1: in Fourier2; P2: in out Fourier2) is
  begin
    P2.T := P1.T;
    Copy(P1.C,P2.C);
    if not STrunc then Copy(P1.E,P2.E); end if;
  end Copy;

  procedure Swap(P1,P2: in out Fourier2) is
    T1: constant FType := P1.T;
  begin
    P1.T := P2.T;
    P2.T := T1;
    Swap(P1.C,P2.C);
    if not STrunc then Swap(P1.E,P2.E); end if;
  end Swap;
   
  procedure BallAt0(R: in Radius; P: in out Fourier2) is
  begin
    SetZero(P.C);
    if not STrunc then
      SetZero(P.E);
      P.E(0,0) := R;
    end if;
  end BallAt0;
   
  function BallAt0(R: Radius) return Fourier2 is
    P: Fourier2;
  begin
    BallAt0(R,P);
    return P;
  end BallAt0;
   
  procedure ToErr(P: in out Fourier2) is
    C: FCoeff renames P.C;
  begin
    for I in Freq1 loop
      for J in Freq2 loop
        ToErr(C(I,J));
      end loop;
    end loop;
  end ToErr;
   
   procedure ToErr(P1: in Fourier2; P2: in out Fourier2) is
      C1: FCoeff renames P1.C;
      E2: FErr renames P2.E;
   begin
      SetZero(P2.C);
      if not STrunc then
      for I in Freq1 loop
	 for J in Freq2 loop
	    E2(I,J) := MaxNorm(C1(I,J));
	 end loop;
      end loop;
      end if;
      P2.T := P1.T;
   end ToErr;
   
   function ToErr(P: Fourier2) return Fourier2 is
      Q: Fourier2;
   begin
      ToErr(P,Q);
      return Q;
   end ToErr;
   
   procedure Center(P: in out Fourier2) is
   begin
      if not STrunc then
   	 Center(P.C);
   	 SetZero(P.E);
      end if;
   end Center;
   
   procedure Center(P1: in Fourier2; P2: in out Fourier2) is
   begin
      Center(P1.C,P2.C);
      if not STrunc then SetZero(P2.E); end if;
      P2.T.P := P1.T.P;
      P2.T.R := Rad;
   end Center;
   
   function Center(P: Fourier2) return Fourier2 is
      Q: Fourier2;
   begin
      Center(P,Q);
      return Q;
   end Center;
   
   procedure ModCenter(P: in out Fourier2) is
   begin
      ModCenter(P.C);
   end ModCenter;
   
   procedure ModCenter(P1: in Fourier2; P2: in out Fourier2) is
   begin
      ModCenter(P1.C,P2.C);
      if not STrunc then Copy(P1.E,P2.E); end if;
      P2.T := P1.T;
   end ModCenter;
   
   function ModCenter(P: Fourier2) return Fourier2 is
      Q: Fourier2;
   begin
      ModCenter(P,Q);
      return Q;
   end ModCenter;
   
   --- addition and multiplication etc.

   procedure Neg(P: in out Fourier2) is
   begin
      Neg(P.C);
   end Neg;

   procedure Neg(P1: in Fourier2; P2: in out Fourier2) is
   begin
      Neg(P1.C,P2.C);
      if not STrunc then Copy(P1.E,P2.E); end if;
      P2.T := P1.T;
   end Neg;

   function "-"(P: Fourier2) return Fourier2 is
      Q: Fourier2;
   begin
      Neg(P.C,Q.C);
      if not STrunc then Copy(P.E,Q.E); end if;
      Q.T := P.T;
      return Q;
   end "-";
   
   procedure Add(I: in Integer; P: in out Fourier2) is
   begin
      Add(I,P.C(0,0));
   end Add;
   
   procedure Add(S: in Scalar; P: in out Fourier2) is
   begin
      Add(S,P.C(0,0));
   end Add;
   
   procedure Add(P1: in Fourier2; P2: in out Fourier2) is
   begin
      if P1.T.P /= P2.T.P then raise Parity_Error with "Fouriers2.Add error"; end if;
      Add(P1.C,P2.C);
      if not STrunc then Add(P1.E,P2.E); end if;
      for J in 1..2 loop
	 if P2.T.R(J)>P1.T.R(J) then P2.T.R(J) := P1.T.R(J); end if;
      end loop;
   end Add;

   procedure Sum(P1,P2: in Fourier2; P3: in out Fourier2) is
   begin
     if P1.T.P /= P2.T.P then raise Parity_Error with "Fouriers2.Sum error"; end if;
     P3.T.P := P2.T.P;
     Sum(P1.C,P2.C,P3.C);
     if not STrunc then Sum(P1.E,P2.E,P3.E); end if;
     for J in 1..2 loop
       P3.T.R(J) := RMin(P1.T.R(J),P2.T.R(J));
     end loop;
  end Sum;

  function "+"(P1,P2: Fourier2) return Fourier2 is
    P3: Fourier2;
  begin
    if P1.T.P /= P2.T.P then raise Parity_Error with "Fouriers2.+ error"; end if;
    P3.T.P := P2.T.P;
    Sum(P1.C,P2.C,P3.C);
    if not STrunc then Sum(P1.E,P2.E,P3.E); end if;
    for J in 1..2 loop
      P3.T.R(J) := RMin(P1.T.R(J),P2.T.R(J));
    end loop;
    return P3;
  end "+";

   procedure Sub(P1: in Fourier2; P2: in out Fourier2) is
   begin
      if P1.T.P /= P2.T.P then raise Parity_Error with "Fouriers2.Sub error"; end if;
      Sub(P1.C,P2.C);
      if not STrunc then Add(P1.E,P2.E); end if;
      for J in 1..2 loop
	 if P2.T.R(J)>P1.T.R(J) then P2.T.R(J) := P1.T.R(J); end if;
      end loop;
   end Sub;

   procedure Diff(P1,P2: in Fourier2; P3: in out Fourier2) is
   begin
      if P1.T.P /= P2.T.P then raise Parity_Error with "Fouriers2.Diff error"; end if;
      P3.T.P := P2.T.P;
      Diff(P1.C,P2.C,P3.C);
      if not STrunc then Sum(P1.E,P2.E,P3.E); end if;
      for J in 1..2 loop
	 P3.T.R(J) := RMin(P1.T.R(J),P2.T.R(J));
      end loop;
   end Diff;

   function "-"(P1,P2: Fourier2) return Fourier2 is
      P3: Fourier2;
   begin
      if P1.T.P /= P2.T.P then raise Parity_Error with "Fouriers2.- error"; end if;
      P3.T.P := P2.T.P;
      Diff(P1.C,P2.C,P3.C);
      if not STrunc then Sum(P1.E,P2.E,P3.E); end if;
      for J in 1..2 loop
	 P3.T.R(J) := RMin(P1.T.R(J),P2.T.R(J));
      end loop;
      return P3;
   end "-";
   
   procedure Mult(R: in Flt; P,Tmp: in out Fourier2) is
      STmp: Scalar renames Tmp.C(0,0);
   begin
      Mult(R,P.C,STmp);
      if not STrunc then Mult(abs(R),P.E); end if;
   end Mult;
   
   procedure Mult(R: in Flt; P: in out Fourier2) is
      STmp: Scalar;
   begin
      Mult(R,P.C,STmp);
      if not STrunc then Mult(abs(R),P.E); end if;
   end Mult;
   
   procedure Prod(R: in Flt; P1: in Fourier2; P2: in out Fourier2) is
   begin
      if R=Zero then
	 SetZero(P1.T,P2);
      else
	 Prod(R,P1.C,P2.C);
	 if not STrunc then Prod(abs(R),P1.E,P2.E); end if;
	 P2.T := P1.T;
      end if;
   end Prod;
   
   function "*"(R: Flt; P: Fourier2) return Fourier2 is
      Q: Fourier2;
   begin
      Prod(R,P,Q);
      return Q;
   end "*";
   
   procedure AddProd(R: in Flt; P1: in Fourier2; P2,Tmp: in out Fourier2) is
      STmp: Scalar renames Tmp.C(0,0);
   begin
      if R /= Zero then
	 AddProd(R,P1.C,P2.C,STmp);
	 if not STrunc then AddProd(abs(R),P1.E,P2.E); end if;
	 for J in 1..2 loop
	    if P2.T.R(J)>P1.T.R(J) then P2.T.R(J) := P1.T.R(J); end if;
	 end loop;
      end if;
   end AddProd;
   
   procedure AddProd(R: in Flt; P1: in Fourier2; P2: in out Fourier2) is
   begin
      if R /= Zero then
	 declare
	    STmp: Scalar;
	 begin
	    AddProd(R,P1.C,P2.C,STmp);
	 end;
	 if not STrunc then AddProd(abs(R),P1.E,P2.E); end if;
	 for J in 1..2 loop
	    if P2.T.R(J)>P1.T.R(J) then P2.T.R(J) := P1.T.R(J); end if;
	 end loop;
      end if;
   end AddProd;
   
   procedure Mult(Q: in Rational; P: in out Fourier2) is
   begin
      if LNum(Q)=0 then
	 SetZero(P);
      else
	 declare
	    Tmp: Scalar;
	 begin
	    Mult(Flt(LNum(Q)),P.C,Tmp);
	    Div(Flt(LDen(Q)),P.C,Tmp);
	 end;
      end if;
   end Mult;
   
  procedure Mult(S: in Scalar; P: in out Fourier2) is
  begin
    Mult(S,P.C);
    if not STrunc then Mult(MaxNorm(S),P.E); end if;
  end Mult;
   
  procedure Prod(S: in Scalar; P1: in Fourier2; P2: in out Fourier2) is
  begin
    if IsZero(S) then
      SetZero(P2);
    else
      Prod(S,P1.C,P2.C);
      if not STrunc then Prod(MaxNorm(S),P1.E,P2.E); end if;
      P2.T := P1.T;
    end if;
  end Prod;
  
  function "*"(S: Scalar; P: Fourier2) return Fourier2 is
    Q: Fourier2;
  begin
    Prod(S,P,Q);
    return Q;
  end "*";
  
  procedure AddProd(S: in Scalar; P1: in Fourier2; P2: in out Fourier2) is
  begin
    if not IsZero(S) then
      AddProd(S,P1.C,P2.C);
      if not STrunc then AddProd(MaxNorm(S),P1.E,P2.E); end if;
      for J in 1..2 loop
	if P2.T.R(J)>P1.T.R(J) then P2.T.R(J) := P1.T.R(J); end if;
      end loop;
    end if;
  end AddProd;
   
   procedure Mult(P1: in Fourier2; P2: in out Fourier2) is
      Tmp: LF_Pointer;
   begin
      Pool.Allocate(Tmp);
      Prod(P1,P2,Tmp.Data);
      Copy(Tmp.Data,P2);
      Pool.Recycle(Tmp);
   end Mult;
   
   --  2*cos(m)*cos(n) =  cos(m-n) + cos(m+n)
   --  2*sin(m)*sin(n) =  cos(m-n) - cos(m+n)
   --  2*cos(m)*sin(n) = -sin(m-n) + sin(m+n)
   --  2*sin(m)*cos(n) =  sin(m-n) + sin(m+n)
   
   procedure Prod(P1,P2: in Fourier2; P3: in out Fourier2) is
   begin
      SetZero(P1.T.P*P2.T.P,P1.T.R,P3);
      AddProd(P1,P2,P3);
   end Prod;
   
   function "*"(P1,P2: ParPair) return ParPair is
   begin
      return ((P1(1)+P2(1)) mod 2,(P1(2)+P2(2)) mod 2);
   end "*";

   procedure Signs(P1,P2: in Parity; Sum,DiffPlus,DiffMinus: out Flt) is
   begin
      Sum := One;
      DiffPlus := One;
      DiffMinus := One;
      if P1=0 then
	 if P2=1 then  DiffPlus := NegOne; end if;
      elsif P2=0 then
	 DiffMinus := NegOne;
      else
	 Sum := NegOne;
      end if;
   end Signs;

   procedure AddProdCC(F1,F2: in Fourier2; F3: in out Fourier2) is
      --- assuming correct NormFacs
      MS,MD,NS,ND: Integer;
      Xs,Ys,Xd,Xdp,Xdm,Yd,Ydp,Ydm,XsYs: Flt;
      C1,C2,C12: Scalar;
      FC1: FCoeff renames F1.C;
      FC2: FCoeff renames F2.C;
      FC3: FCoeff renames F3.C;
      FE3: FErr renames F3.E;
   begin
      Signs(F1.T.P(1),F2.T.P(1),Xs,Xdp,Xdm);
      Signs(F1.T.P(2),F2.T.P(2),Ys,Ydp,Ydm);
      XsYs := Xs*Ys;
      for M1 in Freq1 loop
	 for N1 in Freq2 loop
	    C1 := FC1(M1,N1);
	    if C1 /= SZero then
	       for M2 in Freq1 loop
		  MS := M1+M2;
		  if M2<M1 then
		     MD := M1-M2;
		     Xd := Xdp;
		  else
		     MD := M2-M1;
		     Xd := Xdm;
		  end if;
		  for N2 in Freq2 loop
		     C2 := FC2(M2,N2);
		     if C2 /= SZero then
			NS := N1+N2;
			if N2<N1 then
			   ND := N1-N2;
			   Yd := Ydp;
			else
			   ND := N2-N1;
			   Yd := Ydm;
			end if;
			C12 := Quarter*(C1*C2);       --- factor 1/4
			Add((Xd*Yd)*C12,FC3(MD,ND));
			if MS <= Deg1 then
			   Add((Xs*Yd)*C12,FC3(MS,ND));
			   if NS <= Deg2 then
			      Add(XsYs*C12,FC3(MS,NS));
			   elsif not STrunc then
			      Add(MaxNorm(NormFacs(MS,NS)*C12),FE3(MS,NS));
			   end if;
			elsif not STrunc then
			   Add(MaxNorm(NormFacs(MS,ND)*C12),FE3(MS,ND));
			   Add(MaxNorm(NormFacs(MS,NS)*C12),FE3(MS,NS));
			end if;
			if NS <= Deg2 then
			   Add((Xd*Ys)*C12,FC3(MD,NS));
			elsif not STrunc then
			   Add(MaxNorm(NormFacs(MD,NS)*C12),FE3(MD,NS));
			end if;
		     end if;
		  end loop;
	       end loop;
	    end if;
	 end loop;
      end loop;
   end AddProdCC;

   procedure AddProdCE(F1,F2: in Fourier2; F3: in out Fourier2) is
      --- assuming correct NormFacs
      MS,MD,NS,ND: Integer;
      E1,E2: Flt;
      C1,WMD,WND: Scalar;
      FC1: FCoeff renames F1.C;
      FE2: FErr renames F2.E;
      FE3: FErr renames F3.E;
   begin
      if IsZero(FE2) then return; end if;
      for M1 in Freq1 loop
	 for N1 in Freq2 loop
	    C1 := FC1(M1,N1);
	    if C1 /= SZero then
	       E1 := MaxNorm(Quarter*NormFacs(M1,N1)*C1);  --- factor 1/4
	       for M2 in FreqE1 loop
		  MS := IMin(M2+M1,DegE1);
		  MD := IMax(M2-M1,0);
		  WMD := Sqr(NormFacs1(-IMin(M1,M2)));
		  for N2 in FreqE2 loop
		     E2 := FE2(M2,N2);
		     if E2 /= Zero then
			NS := IMin(N2+N1,DegE2);
			ND := IMax(N2-N1,0);
			WND := Sqr(NormFacs2(-IMin(N1,N2)));
			AddProd(E1,E2,FE3(MS,NS));
			AddProd(Sup(E1*WND),E2,FE3(MS,ND));
			AddProd(Sup(E1*WMD),E2,FE3(MD,NS));
			AddProd(Sup(E1*(WND*WMD)),E2,FE3(MD,ND));
		     end if;
		  end loop;
	       end loop;
	    end if;
	 end loop;
      end loop;
   end AddProdCE;

   procedure AddProdEE(F1,F2: in Fourier2; F3: in out Fourier2) is
      --- assuming correct NormFacs
      MS,NS: Integer;
      E1,E2: Flt;
      WMD,WND: Scalar;
      FE1: FErr renames F1.E;
      FE2: FErr renames F2.E;
      FE3: FErr renames F3.E;
   begin
      if IsZero(FE2) then return; end if;
      for M1 in FreqE1 loop
	 for N1 in FreqE2 loop
	    E1 := FE1(M1,N1);
	    if E1 /= Zero then
	       Mult(Quarter,E1);    --- factor 1/4
	       for M2 in FreqE1 loop
		  MS := IMin(M2+M1,DegE1);
		  WMD := Sqr(NormFacs1(-IMin(M1,M2)));
		  for N2 in FreqE2 loop
		     E2 := FE2(M2,N2);
		     if E2 /= Zero then
			NS := IMin(N2+N1,DegE2);
			WND := Sqr(NormFacs2(-IMin(N1,N2)));
			AddProd(E1,E2,FE3(MS,NS));
			AddProd(Sup(E1*WND),E2,FE3(MS,0));
			AddProd(Sup(E1*WMD),E2,FE3(0,NS));
			AddProd(Sup(E1*(WND*WMD)),E2,FE3(0,0));
		     end if;
		  end loop;
	       end loop;
	    end if;
	 end loop;
      end loop;
   end AddProdEE;
   
   function "*"(P1,P2: Fourier2) return Fourier2 is
      P3: Fourier2;
   begin
      Prod(P1,P2,P3);
      return P3;
   end "*";
   
   procedure Normalize(F: in out Fourier2) is
      P: ParPair renames F.T.P;
      E: FErr renames F.E;
   begin
      if P(1)=1 then
	 ZeroRow(0,F.C);
	 if not STrunc then
	    for N in FreqE2 loop
	       Add(E(0,N),E(1,N));
	       E(0,N) := Zero;
	    end loop;
	 end if;
      end if;
      if P(2)=1 then
	 ZeroColumn(0,F.C);
	 if not STrunc then
	    for M in FreqE1 loop
	       Add(E(M,0),E(M,1));
	       E(M,0) := Zero;
	    end loop;
	 end if;
      end if;
   end Normalize;

   procedure AddProd(P1,P2: in Fourier2; P3: in out Fourier2) is
      T1: FType renames P1.T;
      T2: FType renames P2.T;
      T3: FType renames P3.T;
   begin
      if T1.P*T2.P /= T3.P then
	 raise Parity_Error with "Fouriers2.AddProd parity error";
      end if;
      T3.R := Min(Min(T1.R,T2.R),T3.R);
      UpdateNormFacs(T3.R);
      AddProdCC(P1,P2,P3);
      if not STrunc then
      	 AddProdCE(P1,P2,P3);
      	 AddProdCE(P2,P1,P3);
      	 AddProdEE(P1,P2,P3);
      end if;
      Normalize(P3);
   end AddProd;

   procedure Div(R: in Flt; P,Tmp: in out Fourier2) is
      E: FErr renames P.E;
      A: Flt renames Tmp.E(0,0);
      STmp: Scalar renames Tmp.C(0,0);
   begin
      Div(R,P.C,STmp);
      if not STrunc then
	 A := abs(R);
	 for I in FreqE1 loop
	    for J in FreqE2 loop
	       E(I,J) := E(I,J)/A;
	    end loop;
	 end loop;
      end if;      
   end Div;
   
   procedure Div(R: in Flt; P: in out Fourier2) is
      E: FErr renames P.E;
      A: Flt;
      STmp: Scalar;
   begin
      Div(R,P.C,STmp);
      if not STrunc then
	 A := abs(R);
	 for I in FreqE1 loop
	    for J in FreqE2 loop
	       E(I,J) := E(I,J)/A;
	    end loop;
	 end loop;
      end if;      
   end Div;
   
   procedure Quot(P1: in Fourier2; R: in Flt; P2: in out Fourier2) is
      E1: FErr renames P1.E;
      E2: FErr renames P2.E;
      A: Flt;
   begin
      Quot(P1.C,R,P2.C);
      if not STrunc then
	 A := abs(R);
	 for I in FreqE1 loop
	    for J in FreqE2 loop
	       E2(I,J) := E1(I,J)/A;
	    end loop;
	 end loop;
      end if;
      P2.T := P1.T;
   end Quot;

   function "/"(P: Fourier2; R: Flt) return Fourier2 is
      Q: Fourier2;
   begin
      Quot(P,R,Q);
      return Q;
   end "/";

   procedure Mult(P1: in Fourier2; P2,Tmp: in out Fourier2) is
   begin
      Prod(P1,P2,Tmp);
      Copy(Tmp,P2);
   end Mult;  
   
   function Norm(E: FErr) return Scalar is
      R: Flt := Zero;
   begin
      for M in FreqE1 loop
	 for N in FreqE2 loop
	    Add(E(M,N),R);
	 end loop;
      end loop;
      return BallAt0(R);
   end Norm;
   
   function Norm(R: RadPair; C: FCoeff) return Scalar is
      S: Scalar := SZero;
   begin
      UpdateNormFacs(R);
      for M in Freq1 loop
         for N in Freq2 loop
	    AddProd(NormFacs(M,N),Norm(C(M,N)),S);
	 end loop;
      end loop;
      return S;
   end Norm;

   function Norm(R: RadPair; F: Fourier2) return Scalar is
   begin
      if STrunc then
	 return Norm(R,F.C);
      else
	 return Norm(R,F.C)+Norm(F.E);
      end if;
   end Norm;

   function Norm(F: Fourier2) return Scalar is
   begin
      return Norm(F.T.R,F);
   end Norm;
   
   procedure Norm(F: in Fourier2; Hi,Lo: in out Scalar) is
      Thres: constant Natural := 10;
   begin
      Hi := SZero;
      Lo := SZero;
      UpdateNormFacs(F.T.R);
      for M in Freq1 loop
         for N in Freq2 loop
	    if M+N<Thres then
	       AddProd(NormFacs(M,N),Norm(F.C(M,N)),Lo);
	    else
	       AddProd(NormFacs(M,N),Norm(F.C(M,N)),Hi);
	    end if;
	 end loop;
      end loop;
   end Norm;
   
   procedure Sqr(P: in out Fourier2) is
      Tmp: LF_Pointer;
   begin
      Pool.Allocate(Tmp);
      Prod(P,P,Tmp.Data);
      Copy(Tmp.Data,P);
      Pool.Recycle(Tmp);
   end Sqr;
   
   function Sqr(P: Fourier2) return Fourier2 is
      Q: Fourier2;
   begin
      Prod(P,P,Q);
      return Q;
   end Sqr;
   
   --- conversion and i/o

   function Approx(P: Fourier2) return Flt is
   begin
      return Approx(P.C(0,0));
   end Approx;

   procedure Assign(I: in Integer; P: in out Fourier2) is
   begin
      SetZero(P);
      Assign(I,P.C(0,0));
   end Assign;
   
   procedure Assign(Q: in Rational; P: in out Fourier2) is
   begin
      SetZero(P);
      Assign(Q,P.C(0,0));
   end Assign;
   
   procedure Assign(R: in Flt; P: in out Fourier2) is
   begin
      SetZero(P);
      Assign(R,P.C(0,0));
   end Assign;
   
   function Scal(I: Integer) return Fourier2 is
      P: Fourier2;
   begin
      SetZero(P);
      Assign(I,P.C(0,0));
      return P;
   end Scal;
   
   function Scal(Q: Rational) return Fourier2 is
      P: Fourier2;
   begin
      SetZero(P);
      Assign(Q,P.C(0,0));
      return P;
   end Scal;

   function Scal(R: Flt) return Fourier2 is
      P: Fourier2;
   begin
      SetZero(P);
      Assign(R,P.C(0,0));
      return P;
   end Scal;
   
   function Scal(S: Scalar) return Fourier2 is
      P: Fourier2;
   begin
      SetZero(P);
      Copy(S,P.C(0,0));
      return P;
   end Scal;
   
   procedure Enclose(R1,R2: in Flt; P: in out Fourier2) is
   begin
      SetZero(P);
      Enclose(R1,R2,P.C(0,0));
   end Enclose;
   
   function Enclose(R1,R2: Flt) return Fourier2 is
      P: Fourier2;
   begin
      SetZero(P);
      Enclose(R1,R2,P.C(0,0));
      return P;
   end Enclose;

   procedure Show1(S: in String; T: in FType) is
   begin
      if T.P(1)=0 then
	 Show0(S & " cos*",False);
      else
	 Show0(S & " sin*",False);
      end if;
      if T.P(2)=0 then
	 Show2("cos  Rho",T.R(1),T.R(2));
      else
	 Show2("sin  Rho",T.R(1),T.R(2));
      end if;
      New_Line;
   end Show1;

   procedure Show1(S: in String; F: in Fourier2) is
      W: constant Positive := 4;
   begin
      if IsZero(F) then
	 Put(S & "  Zero_Fourier2");
      else
	 Show1(S,F.T);
	 Show1(S,F.C);
	 if not STrunc then
	    New_Line;
	    for M in FreqE1 loop
	       for N in FreqE2 loop
		  if F.E(M,N) /= Zero then
		     Show1(S & " Err " & Strng(M,W) & Strng(N,W),F.E(M,N));
		  end if;
	       end loop;
	    end loop;
	 end if;
      end if;
      New_Line;
   end Show1;

   procedure Show2(S: in String; F1,F2: in Fourier2) is
      W: constant Positive := 4;
   begin
      if IsZero(F1) and then IsZero(F2) then
	 Put(S & "  Zero_Fouriers2");
      else
	 Show1(S & "1st  ",F1.T);
	 Show1(S & "2nd  ",F2.T);
	 Show2(S,F1.C,F2.C);
	 if not STrunc then
	    New_Line;
	    for M in FreqE1 loop
	       for N in FreqE2 loop
		  if (F1.E(M,N) /= Zero) or else (F2.E(M,N) /= Zero) then
		     Show2(S & " Err " & Strng(M,W) & Strng(N,W),F1.E(M,N),F2.E(M,N));
		  end if;
	       end loop;
	    end loop;
	 end if;
      end if;
      New_Line;
   end Show2;

   function EffDeg1(C: FCoeff) return Integer is
   begin
      for M in reverse Freq1 loop
	 for N in Freq2 loop
	    if C(M,N) /= SZero then return M; end if;
	 end loop;
      end loop;
      return -1;
   end EffDeg1;

   function EffDeg2(C: FCoeff) return Integer is
   begin
      for N in reverse Freq2 loop
	 for M in Freq1 loop
	    if C(M,N) /= SZero then return N; end if;
	 end loop;
      end loop;
      return -1;
   end EffDeg2;

   procedure EffDegs0(C: in FCoeff; D1,D2: out Integer) is
   begin
      D1 := IMax(EffDeg1(C),0);
      D2 := IMax(EffDeg2(C),0);
   end EffDegs0;

   function EffDeg1(E: FErr) return Integer is
   begin
      for M in reverse Freq1 loop
	 for N in Freq2 loop
	    if E(M,N) /= Zero then return M; end if;
	 end loop;
      end loop;
      return -1;
   end EffDeg1;

   function EffDeg2(E: FErr) return Integer is
   begin
      for N in reverse Freq2 loop
	 for M in Freq1 loop
	    if E(M,N) /= Zero then return N; end if;
	 end loop;
      end loop;
      return -1;
   end EffDeg2;

   procedure EffDegs0(E: in FErr; D1,D2: out Integer) is
   begin
      D1 := IMax(0,EffDeg1(E));
      D2 := IMax(0,EffDeg2(E));
   end EffDegs0;

   procedure Put(F: in File_Type; H: in Fourier2; Decimal: in Boolean := False) is
      D1,D2: Integer;
   begin
      Put(F,H.T.P(1));
      Put(F,H.T.P(2));
      Put(F,H.T.R(1),Decimal);
      Put(F,H.T.R(2),Decimal);
      EffDegs0(H.C,D1,D2);
      Put(F,D1);
      Put(F,D2);
      for M in 0 .. D1 loop
	 for N in 0 .. D2 loop
	    Put(F,H.C(M,N),Decimal);
	 end loop;
      end loop;
      EffDegs0(H.E,D1,D2);
      Put(F,D1);
      Put(F,D2);
      for M in 0 .. D1 loop
	 for N in 0 .. D2 loop
	    Put(F,H.E(M,N),Decimal);
	 end loop;
      end loop;
   end Put;

   procedure Get(F: in File_Type; H: in out Fourier2; Decimal: in Boolean := False) is
      D1,D2: Integer := 0;
      E: Flt := Zero;
      C: Scalar;
   begin
      SetZero(H);
      Get(F,H.T.P(1));
      Get(F,H.T.P(2));
      Get(F,H.T.R(1),Decimal);
      Get(F,H.T.R(2),Decimal);
      Get(F,D1);
      Get(F,D2);
      for M in 0 .. D1 loop
	 for N in 0 .. D2 loop
	    Get(F,C,Decimal);
	    AddCoeff(M,N,C,H);
	 end loop;
      end loop;
      Get(F,D1);
      Get(F,D2);
      for M in 0 .. D1 loop
	 for N in 0 .. D2 loop
	    Get(F,E,Decimal);
	    AddBall(M,N,E,H.E);
	 end loop;
      end loop;
   end Get;

   procedure Write(Name: in String; H: in Fourier2; Decimal: in Boolean := False) is
      F: File_Type;
   begin
      if Verbosity>1 then Put_Line("Writing " & Name); end if;
      Create(F,Out_File,Name);
      Put(F,H,Decimal);
      Close(F);
   end Write;

   procedure Read(Name: in String; H: in out Fourier2; Decimal: in Boolean := False) is
      F: File_Type;
   begin
      if Verbosity>1 then Put_Line("Reading " & Name); end if;
      Open(F,In_File,Name);
      Get(F,H,Decimal);
      Close(F);
   end Read;
   
  procedure NumRead(Name: in String; H: in out Fourier2; Decimal: in Boolean := False) is
    C: FCoeff renames H.C;
  begin
    Read(Name,H,Decimal);
    if not STrunc then
      for I in C'Range(1) loop
        for J in C'Range(2) loop
	  Center(C(I,J));
        end loop;
      end loop;
      SetZero(H.E);
     end if;
  end NumRead;
  
  procedure CheckType11(F: in Fourier2; C11: out Scalar; OK: out Boolean) is
--- far from optimal
    C: FCoeff renames F.C;
    E: FErr renames F.E;
    R: Radius := Zero;
    A11: Radius;
  begin
    if not STrunc then
      for I in reverse FreqE1 loop
        for J in reverse FreqE2 loop
          R := R+E(I,J);
        end loop;
      end loop;
    end if;
    C11 := C(1,1)+DiskAt0(R);
    A11 := Inf(abs(C(1,1)+DiskAt0(Two*R)));
    OK := False;
    for I in Freq1 loop
      for J in Freq2 loop
        if (I /= 1) or else (J /= 1) then
          if A11<MaxNorm(C(I,J)) then return; end if;
        end if;
      end loop;
    end loop;
    OK := True;
  end CheckType11;
  
   procedure Plot(G: in Fourier2; L: Positive := 50) is
      use Flt_IO;
      F: File_Type;
      P: constant Flt := Approx(Pi);
      X,Y,Z: Scalar;
   begin
      Create(F,Out_File,"out.dat");
      for M in 0 .. L loop
	 for N in 0 .. L loop
	    X := P*Scal(M)/Scal(L);
	    Y := P*Scal(N)/Scal(L);
	    Z := Val(X,Y,G);
	    Put(F,Approx(X),3,3,3);
	    Put(F,Approx(Y),3,3,3);
	    Put(F,Approx(Z),3,3,3);
	    New_Line(F);
	 end loop;
      end loop;
      Close(F);
   end Plot;

  function GuessRho1(F: Fourier2) return Flt is
    use Flt_EF;
    C: FCoeff renames F.C;
    D1: constant Integer := EffDeg1(C);
    M0: constant Integer := (2*D1)/3;
    R,FM,T: Flt := Zero;
  begin
    for M in M0 .. D1 loop
      FM := Flt(M);
      for N in 0 .. Deg2 loop
        T := MaxNorm(C(M,N));
        if T /= Zero then
          R := RMax(Exp(Log(T)/FM),R);
        end if;
      end loop;
    end loop;
    return One/R;
  end GuessRho1;
  
  function GuessRho2(F: Fourier2) return Flt is
    use Flt_EF;
    C: FCoeff renames F.C;
    D2: constant Integer := EffDeg2(C);
    N0: constant Integer := (2*D2)/3;
    R,FN,T: Flt := Zero;
  begin
    for N in N0 .. D2 loop
      FN := Flt(N);
      for M in 0 .. Deg1 loop
        T := MaxNorm(C(M,N));
        if T /= Zero then
          R := RMax(Exp(Log(T)/FN),R);
        end if;
      end loop;
    end loop;
    return One/R;
  end GuessRho2;
  
  function GuessRho(F: Fourier2) return RadPair is
  begin
    return (GuessRho1(F),GuessRho2(F));
  end GuessRho;
  
  procedure GuessRho(F: in Fourier2) is
    T: FType;
  begin
    T.P := F.T.P;
    T.R := GuessRho(F);
    Show1("guessed domain:",T);
  end GuessRho;
  
--- misc

   function Get_Precision(P: Fourier2) return Positive is
   begin
      return Get_Precision(P.C(0,0));
   end Get_Precision;
   
   procedure Proper_Rounding(Dummy: in Fourier2) is
      pragma Unreferenced (Dummy);
   begin
      Proper_Rounding(SZero);
   end Proper_Rounding;
   
   procedure Free_Cache(Dummy: in Fourier2) is
      pragma Unreferenced (Dummy);
   begin
      Pool.Free_All;
      Free_Cache(SZero);
   end Free_Cache;
   
   --------------- other standard procedures   
   
   function UpInvNormFac(R: Radius; I: Integer) return Radius is
      PowR,PowQ: Flt := NegHalf;  
   begin
      for K in 1 .. abs(I) loop
	 PowR := PowR*R;
	 PowQ := PowQ/R;
      end loop;
      return One/(-(PowR+PowQ));
   end UpInvNormFac;
   
   function NumCoeffs(D1,D2: Natural; JustOdd: Boolean) return Positive is
---- number of coefficients up to degrees D1,D2  
   begin
     if JustOdd then
       if ((D1*D2) mod 2)=0 then 
         raise Parity_Error with "Fouriers2.NumCoeffs: expect odd degrees";
       end if;
       return (D1/2+1)*(D2/2+1);
     else
       return (D1+1)*(D2+1);
     end if;
   end NumCoeffs;
   
   function Coeff(M,N: Natural; R: RadPair; E: FErr) return Scalar is
      --- too slow for use in loops
      M0: constant Natural := IMin(M,DegE1);
      N0: constant Natural := IMin(N,DegE2);
      S: Flt := Zero;
   begin
      for M1 in 0 .. M0 loop
	 for N1 in 0 .. N0 loop
	    Add(E(M1,N1),S);
	 end loop;
      end loop;
      if S=Zero then
	 return BallAt0(S);
      else
	 UpdateNormFacs(R);
	 return Inv(NormFacs(M,N))*BallAt0(S);
      end if;
   end Coeff;

   function Coeff(M,N: Natural; F: Fourier2) return Scalar is
      --- too slow for use in loops
      C,E: Scalar := SZero;
   begin
      if (M <= Deg1) and then (N <= Deg2) then C := F.C(M,N); end if;
      if not STrunc then E := Coeff(M,N,F.T.R,F.E); end if;
      return C+E;
   end Coeff;

   procedure AddBall(M,N: in Natural; R: in Flt; E: in out FErr) is
   begin
      Add(abs(R),E(IMin(M,DegE1),IMin(N,DegE2)));
   end AddBall;

   procedure AddCoeff(M,N: in Natural; S: in Scalar; F: in out Fourier2) is
   begin
      if (M <= Deg1) and then (N <= Deg2) then
	 Add(S,F.C(M,N));
      elsif not (STrunc or else IsZero(S)) then
	 UpdateNormFacs(F.T.R);
	 AddBall(M,N,NormFacs(M,N)*S,F);
      end if;
   end AddCoeff;

   procedure AddBall(M,N: in Natural; S: in Scalar; F: in out Fourier2) is
   begin
      if not STrunc then
	 AddBall(M,N,MaxNorm(S),F.E);
      end if;
   end AddBall;

   procedure AddBall(S: in Scalar; F: in out Fourier2) is
   begin
      if not STrunc then
	 Add(MaxNorm(S),F.E(0,0));
      end if;
   end AddBall;
   
   procedure AddBall(R: in Flt; F: in out Fourier2) is
   begin
      if not STrunc then
	 Add(abs(R),F.E(0,0));
      end if;
   end AddBall;
   
   function UpNormFac(R: Radius; I: Integer) return Radius is
      Q: constant Flt := One/R;
      PowR,PowQ: Flt := Half;  
   begin
      for K in 1 .. abs(I) loop
	 PowR := PowR*R;
	 PowQ := PowQ*Q;
      end loop;
      return PowR+PowQ;
   end UpNormFac;
   
   function Val(X,Y: Scalar; F: Fourier2) return Scalar is
      P1: constant Parity := F.T.P(1);
      P2: constant Parity := F.T.P(2);
      V: Scalar := SZero;
      S: Scalar;
      C: FCoeff renames F.C;
   begin
      for M in Freq1 loop
	 S := CosSin(P1,Flt(M)*X);
	 for N in Freq2 loop
	    V := V+C(M,N)*S*CosSin(P2,Flt(N)*Y);
	 end  loop;
      end loop;
      if not STrunc then V := V+Norm(F.E); end if;
      return V;
   end  Val;
   
begin
  
  UpdateNormFacs((Zero,Zero));
   
end Fouriers2;
