with Ada.Text_IO, Logicals, Globals, Ints, Rationals, Flts, Flt_Matrices;
use Ada.Text_IO, Logicals, Globals, Ints, Rationals, Flts;
with Recycling, Matrices, Vectors, Matrices.More, Vectors.More;

pragma Elaborate_All (Ada.Text_IO,Logicals,Globals,Ints,Rationals,Flts);
pragma Elaborate_All (Recycling,Matrices,Vectors,Matrices.More,Vectors.More);

generic

  Deg1,Deg2: in Positive;
  
-------------------- begin_include scalars Scalar 

  type Scalar is private;

--- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsNumbers(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

--- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>; -- S1 constains S2
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure DiskAt0(R: in Flt; S: in out Scalar) is <>;
  with function DiskAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure CenterDisk(S: in out Scalar) is <>;
  with procedure CenterDisk(S1: in Scalar; S2: in out Scalar) is <>;
  with function CenterDisk(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ModNumbers(S: in out Scalar) is <>;
  with procedure ModNumbers(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModNumbers(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

--- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part
  with function LFloor(S: Scalar) return LInt is <>;
  with function LCeiling(S: Scalar) return LInt is <>;

--- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

--- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with function Pi return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

--- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

--- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  with procedure Proper_Rounding(Dummy: in Scalar) is <>;
  with procedure Free_Cache(S: in Scalar) is <>;
-------------------- end_include scalars 
  
  with package SM is new Matrices (Scalar => Scalar);
  with package SV is new Vectors (Scalar => Scalar, Component => Scalar, SM => SM);

package Fouriers2 is  -- Fourier series in 2d with non-constant coefficients
  
  pragma Elaborate_Body;
  
  package SV_More is new SV.More;
  package SM_More is new SM.More;
  
  DegE1: constant Positive := 2*Deg1;
  DegE2: constant Positive := 2*Deg2;

  subtype Freq1  is Integer range 0 .. Deg1;
  subtype Freq2  is Integer range 0 .. Deg2;
  subtype FreqE1 is Integer range 0 .. DegE1;
  subtype FreqE2 is Integer range 0 .. DegE2;

  subtype FCoeff is SM.Matrix(Freq1,Freq2);
  subtype FErr is  Flt_Matrices.Matrix(FreqE1,FreqE2);  -- assumes rounding up

  type FType is
    record
      P: ParPair;  --- even (0) or odd (1)
      R: RadPair;  --- domain parameter
    end record;

  type Fourier2 is
    record
      T: FType;
      C: FCoeff;
      E: FErr;
    end record;
  
  Rad: RadPair := (Two,Two);     -- default domain
  
--------------- standard Scalar procedures

--- basic
  function Info(Dummy: Fourier2) return Scalar_Info;
  function IsPoly(P: Fourier2) return Boolean;
  function IsSharp(P: Fourier2) return Boolean;         --- polynomial with sharp coeffs
  function IsZero(P: Fourier2) return Boolean;
  function "="(P1,P2: Fourier2) return Boolean;
  procedure SetZero(P: in out Fourier2);
  procedure SetZero(T: in FType; F: in out Fourier2);
  procedure SetZero(Template: in Fourier2; F: in out Fourier2);
  procedure SetZero(P: in ParPair; R: in RadPair; F: in out Fourier2);
  procedure Copy(P1: in Fourier2; P2: in out Fourier2);
  procedure Swap(P1,P2: in out Fourier2);

--- sets
  procedure BallAt0(R: in Radius; P: in out Fourier2);       --- 0-centered ball of radius R
  function BallAt0(R: Radius) return Fourier2;               --- 0-centered ball of radius R
  procedure ToErr(P: in out Fourier2);                    --- 0-centered set containing P
  procedure ToErr(P1: in Fourier2; P2: in out Fourier2);  --- 0-centered set containing P1
  function ToErr(P: Fourier2) return Fourier2;            --- 0-centered set containing P
  procedure Center(P: in out Fourier2);
  procedure Center(P1: in Fourier2; P2: in out Fourier2);
  function Center(P: Fourier2) return Fourier2;
  procedure ModCenter(P: in out Fourier2);
  procedure ModCenter(P1: in Fourier2; P2: in out Fourier2);
  function ModCenter(P: Fourier2) return Fourier2;            --- zero center
  
--- addition and multiplication etc.
  procedure Neg(P: in out Fourier2);
  procedure Neg(P1: in Fourier2; P2: in out Fourier2);
  function "-"(P: Fourier2) return Fourier2;
  procedure Add(I: in Integer; P: in out Fourier2);
  procedure Add(S: in Scalar; P: in out Fourier2);
  procedure Add(P1: in Fourier2; P2: in out Fourier2);
  procedure Sum(P1,P2: in Fourier2; P3: in out Fourier2);
  function "+"(P1,P2: Fourier2) return Fourier2;
  procedure Sub(P1: in Fourier2; P2: in out Fourier2);
  procedure Diff(P1,P2: in Fourier2; P3: in out Fourier2);
  function "-"(P1,P2: Fourier2) return Fourier2;
  procedure Mult(R: in Flt; P,Tmp: in out Fourier2);
  procedure Mult(R: in Flt; P: in out Fourier2);
  procedure Prod(R: in Flt; P1: in Fourier2; P2: in out Fourier2);
  function "*"(R: Flt; P: Fourier2) return Fourier2;
  procedure AddProd(R: in Flt; P1: in Fourier2; P2,Tmp: in out Fourier2);
  procedure AddProd(R: in Flt; P1: in Fourier2; P2: in out Fourier2);
  procedure Mult(Q: in Rational; P: in out Fourier2);
  procedure Mult(S: in Scalar; P: in out Fourier2);
  procedure Prod(S: in Scalar; P1: in Fourier2; P2: in out Fourier2);
  function "*"(S: Scalar; P: Fourier2) return Fourier2;
  procedure AddProd(S: in Scalar; P1: in Fourier2; P2: in out Fourier2);
  procedure Mult(P1: in Fourier2; P2: in out Fourier2);
  function "*"(P1,P2: ParPair) return ParPair;
  procedure Prod(P1,P2: in Fourier2; P3: in out Fourier2);
  function "*"(P1,P2: Fourier2) return Fourier2;
  procedure Normalize(F: in out Fourier2);
  procedure AddProd(P1,P2: in Fourier2; P3: in out Fourier2);
  procedure Div(R: in Flt; P,Tmp: in out Fourier2);
  procedure Div(R: in Flt; P: in out Fourier2);
  procedure Quot(P1: in Fourier2; R: in Flt; P2: in out Fourier2);
  function "/"(P: Fourier2; R: Flt) return Fourier2;

--- functions
  function Norm(R: RadPair; C: FCoeff) return Scalar;
  function Norm(E: FErr) return Scalar;
  function Norm(R: RadPair; F: Fourier2) return Scalar;
  function Norm(F: Fourier2) return Scalar;
  procedure Norm(F: in Fourier2; Hi,Lo: in out Scalar);
  procedure Sqr(P: in out Fourier2);
  function Sqr(P: Fourier2) return Fourier2;
   
--- conversion and i/o
  function Approx(P: Fourier2) return Flt;
  procedure Assign(I: in Integer; P: in out Fourier2);
  procedure Assign(Q: in Rational; P: in out Fourier2);
  procedure Assign(R: in Flt; P: in out Fourier2);
  function Scal(I: Integer) return Fourier2;
  function Scal(Q: Rational) return Fourier2;
  function Scal(R: Flt) return Fourier2;
  function Scal(S: Scalar) return Fourier2;
  procedure Enclose(R1,R2: in Flt; P: in out Fourier2);
  function Enclose(R1,R2: Flt) return Fourier2;
  procedure Show1(S: in String; T: in FType);
  procedure Show1(S: in String; F: in Fourier2);
  procedure Show2(S: in String; F1,F2: in Fourier2);
  procedure Put(F: in File_Type; H: in Fourier2; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; H: in out Fourier2; Decimal: in Boolean := False);
  procedure Write(Name: in String; H: in Fourier2; Decimal: in Boolean := False);
  procedure Read(Name: in String; H: in out Fourier2; Decimal: in Boolean := False);
  
--- more
  procedure NumRead(Name: in String; H: in out Fourier2; Decimal: in Boolean := False);
  procedure CheckType11(F: in Fourier2; C11: out Scalar; OK: out Boolean);
  procedure Plot(G: in Fourier2; L: Positive := 50);
  function GuessRho1(F: Fourier2) return Flt;
  function GuessRho2(F: Fourier2) return Flt;
  function GuessRho(F: Fourier2) return RadPair;
  procedure GuessRho(F: in Fourier2);

--- misc
  function Get_Precision(P: Fourier2) return Positive;
  procedure Proper_Rounding(Dummy: in Fourier2);
  procedure Free_Cache(Dummy: in Fourier2);
  
--------------- other standard procedures
  function NumCoeffs(D1,D2: Natural; JustOdd: Boolean) return Positive;
  function Coeff(M,N: Natural; F: Fourier2) return Scalar;
  procedure AddCoeff(M,N: in Natural; S: in Scalar; F: in out Fourier2);
  procedure AddBall(M,N: in Natural; S: in Scalar; F: in out Fourier2);
  procedure AddBall(S: in Scalar; F: in out Fourier2);
  procedure AddBall(R: in Flt; F: in out Fourier2);
  procedure AddBall(M,N: in Natural; R: in Flt; E: in out FErr);
  function Val(X,Y: Scalar; F: Fourier2) return Scalar;
  
  type Fourier2_Access is access Fourier2;
  
private
  
  procedure UpdateNormFacs(R: in RadPair);
  function NotExp(X: Flt) return Scalar;
  procedure Powers(S: in Scalar; W: in out SV.Vector);
  function CosSin(P: Parity; S: Scalar) return Scalar;
  procedure AddProd(S: in Scalar; C1: in FCoeff; C2: in out FCoeff; Tmp: in out Scalar);
  procedure SubProd(S: in Scalar; C1: in FCoeff; C2: in out FCoeff; Tmp: in out Scalar);
  procedure Signs(P1,P2: in Parity; Sum,DiffPlus,DiffMinus: out Flt);
  procedure AddProdCC(F1,F2: in Fourier2; F3: in out Fourier2);
  procedure AddProdCE(F1,F2: in Fourier2; F3: in out Fourier2);
  procedure AddProdEE(F1,F2: in Fourier2; F3: in out Fourier2);
  procedure Mult(P1: in Fourier2; P2,Tmp: in out Fourier2);
  function UpInvNormFac(R: Radius; I: Integer) return Radius;
  function UpNormFac(R: Radius; I: Integer) return Radius;
  function EffDeg1(C: FCoeff) return Integer;
  function EffDeg2(C: FCoeff) return Integer;
  procedure EffDegs0(C: in FCoeff; D1,D2: out Integer);
  function EffDeg1(E: FErr) return Integer;
  function EffDeg2(E: FErr) return Integer;
  procedure EffDegs0(E: in FErr; D1,D2: out Integer);
  
  DegE:    constant Positive    := IMax(DegE1,DegE2);
  SZero:   constant Scalar      := Scal(0);
  SHalf:   constant Scalar      := Scal(Half);
  SOne:    constant Scalar      := Scal(1);
  SThree:  constant Scalar      := Scal(3);
  SInfo:   constant Scalar_Info := Info(SZero);
  STrunc:  constant Boolean     := SInfo.IsNumeric;
  Std:     constant Boolean     := SInfo.IsStandard;
  
  LastPair: RadPair := (One,One);
  NormFacs1,NormFacs2: SV.Vector(-DegE .. DegE);
  NormFacs: SM.Matrix(FreqE1,FreqE2);

  FInfo: constant Scalar_Info := (RepType    => SInfo.RepType,
                                  IsStandard => SInfo.IsStandard,
                                  IsNumeric  => SInfo.IsNumeric,
                                  IsComplex  => SInfo.IsComplex,
                                  NReps      => SInfo.NReps*(2*(DegE1+DegE2)+2),
                                  RMode      => SInfo.RMode);
  
  package Alloc is new Recycling (Item => Fourier2);
  subtype LF_Pointer is Alloc.LD_Pointer;
  Pool: Alloc.Protected_Data;
  
end Fouriers2;
