
package Globals is

  type Rounding_Direction is (Up, Down, Nearest, Truncate);
  type Rounding_Mask is (Strict, Relaxed);

  type Rounding_Mode is
    record
      Direction: Rounding_Direction := Nearest;
      Mask: Rounding_Mask := Relaxed;
    end record;
  
  type Scalar_Info is
    record
      RepType:    String(1..10);
      IsStandard: Boolean := True;
      IsNumeric:  Boolean := True;
      IsComplex:  Boolean := False;
      NReps:      Natural := 1;
      RMode:      Rounding_Mode;
    end record;
  
  Initial_Precision: constant Positive := 512;    --- used with MPFR during elaboration
  
  CheckRange: Boolean := False; -- for matrices and vectors
  Verbosity:  Natural  := 0;    -- 0: just error messages and important warnings
                                -- 1: also high level program info
                                -- 2: also TraceEnter and TraceLeave
                                -- 3: also domain errors with info in numeric programs
                                -- 4: more stuff
                                -- 5: almost everything
                                -- 6: insane

  Domain_Error, Dimension_Error, Undefined, Not_Implemented, Sorry, Parity_Error: exception;
  
  NTasks: Positive     := 1;      --- default value
  Stack_Size: Positive := 196608; --- default value
  
  Input_Dir:  constant String := "input/";
  Output_Dir: constant String := "output/";
  Tmp_Dir:    constant String := "tmpdata/";
  
end Globals;
