with Ada.Command_Line, Strings, Ints, Globals;
use Ada.Command_Line, Globals, Ints;

pragma Elaborate_All (Ada.Command_Line,Strings,Ints,Globals);

package body Messages is

  package Txt_IO renames Ada.Text_IO;
  
  procedure Check_True(B: in Boolean) is
  begin
    if B then
      Txt_IO.Put_Line("Check passed");
    else
      Txt_IO.Put_Line("Check failed");
      raise Sorry;
    end if;
    New_Line;
  end Check_True;

  procedure Show(S: in String; NewLine: in Boolean := True) is
    use Strings,Tasks;
  begin
    if Verbosity=0 then return; end if;
    declare
      P: constant File_Access := Task_Log;
    begin
      Put(P,S);
      if NewLine then Flush_New_Line(P); end if;
    end;
  end Show;

  procedure Show(S: in String; R: in Flt; NewLine: in Boolean := True) is
    use Strings,Tasks;
  begin
    if Verbosity=0 then return; end if;
    declare
      P: constant File_Access := Task_Log;
    begin
      Put(P,S);
      Put(P,R,3,DefaultAft,DefaultExp);
      if NewLine then Flush_New_Line(P); end if;
    end;
  end Show;

  procedure Show(S: in String; R1,R2: in Flt; NewLine: in Boolean := True) is
    use Strings,Tasks;
  begin
    if Verbosity=0 then return; end if;
    declare
      P: constant File_Access := Task_Log;
    begin
      Put(P,S);
      Put(P,R1,3,DefaultAft,DefaultExp);
      Put(P,R2,3,DefaultAft,DefaultExp);
      if NewLine then Flush_New_Line(P); end if;
    end;
  end Show;
  
  function GetArg return Boolean is
    S: constant String := GetArg;
  begin
    if (S="True") or else (S="true") then return True; end if;
    if (S="False") or else (S="false") then return False; end if;
    Txt_IO.Put_Line("Messages.GetArg: non-Boolean value");
    raise Sorry;
    return False;
  end GetArg;

  function GetArg return Integer is
    M,L: Integer;
  begin
    Int_IO.Get(Argument(ArgIndex),M,L);
    ArgIndex := ArgIndex+1;
    return M;
  end GetArg;
  
  function GetArg return Character is
    S: constant String := Argument(ArgIndex);
  begin
    ArgIndex := ArgIndex+1;
    return S(S'First);
  end GetArg;
  
  function GetArg return String is
    S: constant String := Argument(ArgIndex);
  begin
    ArgIndex := ArgIndex+1;
    return S;
  end GetArg;

  function GetArg return Flt is
--- careful, input may get rounded
    L: Integer;
    R: Flt;
  begin
    Flt_IO.Get(Argument(ArgIndex),R,L);
    ArgIndex := ArgIndex+1;
    return R;
  end GetArg;
  
  function Age(T: in Timer := Default_Timer) return Flt is
--- age in seconds
  begin
    return Flt(To_Duration(Clock-T));
  end Age;

  function Seconds(T: Timer := Default_Timer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T)));
  end Seconds;

  function Minutes(T: Timer := Default_Timer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T)/60.0));
  end Minutes;

  function Hours(T: Timer := Default_Timer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T)/3600.0));
  end Hours;

  procedure Show_Time(P: in File_Access; T: in Timer := Default_Timer) is
    use Tasks;
    S: constant Duration := To_Duration(Clock-T);
    M: constant Integer := Integer(S)/60;
    H: constant Integer := M/60;
  begin
    if P=null then
      Int_IO.Put(H,3);
      Txt_IO.Put(":");
      Int_IO.Put(M mod 60,2);
      Txt_IO.Put(":");
      Dur_IO.Put(S-Duration(60*M),2,3);
      New_Line;
    else
      declare
        F: File_Type renames P.all;
      begin
        Int_IO.Put(F,H,3);
        Txt_IO.Put(F,":");
        Int_IO.Put(F,M mod 60,2);
        Txt_IO.Put(F,":");
        Dur_IO.Put(F,S-Duration(60*M),2,3);
        New_Line(F);
--        Flush(F);
      end;
    end if;
  end Show_Time;
  
  procedure Show_Time(T: in Timer := Default_Timer) is
    use Tasks;
  begin
    Show_Time(Task_Log,T);
  end Show_Time;
  
  function Starting(Where: String) return Tracer is
    use Tasks;
    H: constant Attribute_Handle := Reference;
    P: constant File_Access := Task_Log(H);
    Level: Natural renames H.Level;
  begin
    if Verbosity>1 then
      if P=null then
        for K in 1 .. Level loop Txt_IO.Put(". "); end loop;
        Txt_IO.Put_Line("[ " & Where);
      else
        declare
          F: File_Type renames P.all;
        begin
          for K in 1 .. Level loop Txt_IO.Put(F,". "); end loop;
          Txt_IO.Put_Line(F,"[ " & Where);
--          Flush(F);
        end;
      end if;
    end if;
    Level := Level+1;
    return (Controlled with T => Clock, H => H, Wait => Adjustements);
  end Starting;      
  
  procedure Adjust(T: in out Tracer) is
--- just count the number of adjustements
  begin
    T.Wait := T.Wait-1;
  end Adjust;
  
  procedure Finalize(T: in out Tracer) is
    use Tasks;
  begin
    if T.Wait /= 0 then  -- T is just being copied to/from stack
      return;            -- wait until procedure actually starts
    end if;
    declare
      Level: Natural renames T.H.Level;
    begin
      Level := Level-1;
      if Verbosity>1 then
        declare
          P: constant File_Access := Task_Log(T.H);
        begin
          if P=null then
            for K in 1 .. Level loop Txt_IO.Put(". "); end loop;
            Txt_IO.Put("] ");
          else
            declare
              F: File_Type renames P.all;
            begin
              for K in 1 .. Level loop Txt_IO.Put(F,". "); end loop;
              Txt_IO.Put(F,"] ");
            end;
          end if;
          Show_Time(P,T.T);
        end;
      end if;
    end;
  end Finalize;
  
  function Seconds(T: Tracer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T.T)));
  end Seconds;
  
  function Minutes(T: Tracer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T.T)/60.0));
  end Minutes;

  function Hours(T: Tracer) return Integer is
  begin
    return Integer(Flt'Floor(Age(T.T)/3600.0));
  end Hours;
    
  procedure Show_Time(T: in Tracer) is
    use Tasks;
  begin
    Show_Time(Task_Log,T.T);
  end Show_Time;
  
end Messages;
