with Ada.Text_IO, Ada.Finalization, Ada.Real_Time, Tasks, Flts;
use Ada.Text_IO, Ada.Finalization, Ada.Real_Time, Flts;

pragma Elaborate_All (Ada.Text_IO,Ada.Finalization,Ada.Real_Time,Tasks,Flts);

package Messages is

  pragma Elaborate_Body;

  procedure Check_True(B: in Boolean);
  procedure Show(S: in String; NewLine: in Boolean := True);
  procedure Show(S: in String; R: in Flt; NewLine: in Boolean := True);
  procedure Show(S: in String; R1,R2: in Flt; NewLine: in Boolean := True);
  
--- get next command line argument
  function GetArg return Boolean;
  function GetArg return Integer;
  function GetArg return Character;
  function GetArg return String;
  function GetArg return Flt;      -- caution, input may get rounded

  subtype Timer is Time;
  Default_Timer: Timer := Clock;

  function Reset_Timer return Timer renames Clock;
  function Age(T: Timer := Default_Timer) return Flt;
  function Seconds(T: Timer := Default_Timer) return Integer;
  function Minutes(T: Timer := Default_Timer) return Integer;
  function Hours(T: Timer := Default_Timer) return Integer;
  procedure Show_Time(T: in Timer := Default_Timer);             --- H:M:S

  type Tracer is new Controlled with private;
  
  function Starting(Where: String) return Tracer;
  function Seconds(T: Tracer) return Integer;
  function Minutes(T: Tracer) return Integer;
  function Hours(T: Tracer) return Integer;
  procedure Show_Time(T: in Tracer);

  ArgIndex:   Positive := 1;
  DefaultAft: Positive := 3;
  DefaultExp: Positive := 3;

private

  package Dur_IO is new Ada.Text_IO.Fixed_IO(Duration);
  
  type Tracer is new Controlled with
    record
      H: Tasks.Attribute_Handle;       -- task attributes
      T: Timer;
      Wait: Integer;
    end record;
  
  Adjustements: constant Integer := 2; -- may be compiler dependent
  
  procedure Show_Time(P: in File_Access; T: in Timer := Default_Timer);
  overriding procedure Adjust(T: in out Tracer);
  overriding procedure Finalize(T: in out Tracer);
  
end Messages;
