
package body Rationals is
  
  function GCD(N,D: LInt) return LInt is
    AN: LInt := abs(N);
    AD: LInt := abs(D);
    Tmp: LInt;
  begin
    while (AN /= 0) loop
      Tmp  := AN;
      AN := AD mod Tmp;
      AD := Tmp;
    end loop;
    if AD=0 then raise Constraint_Error; end if;
    return AD;
  end GCD;
  
  function Width(I: LInt; Base: Positive := 10) return Positive is
    B: constant LInt := LInt(Base);
    L: Integer;
    K: LInt;
  begin
    K := abs(I);
    if I<0 then L := 2; else L := 1; end if;
    while K >= B loop
      K := K/B;
      L := L+1;
    end loop;
    return L;
  end Width;
  
  function LRat(N,D: LInt) return Rational is
    G: constant LInt := GCD(N,D);
  begin
    if D>0 then
      return (N/G,D/G);
    end if;
    if D=0 then raise Constraint_Error; end if;
    return ((-N)/G,(-D)/G);
  end LRat;
  
  function IRat(N,D: Integer) return Rational is
  begin
    return Rat(Lint(N),Lint(D));
  end IRat;
  
  function LNum(Q: Rational) return Lint is
  begin
    return Q.Num;
  end LNum;
  
  function INum(Q: Rational) return Integer is
  begin
    return Integer(Q.Num);
  end INum;
  
  function LDen(Q: Rational) return Lint is
  begin
    return Q.Den;
  end LDen;
  
  function IDen(Q: Rational) return Integer is
  begin
    return Integer(Q.Den);
  end IDen;
  
  function "-"(Q: Rational) return Rational is
  begin
    return (-Q.Num,Q.Den);
  end "-";
  
  function "+"(I: LInt; Q: Rational) return Rational is
  begin
    return (I*Q.Den+Q.Num)/Q.Den;
  end "+";
  
  function "+"(Q: Rational; I: LInt) return Rational is
  begin
    return (Q.Num+I*Q.Den)/Q.Den;
  end "+";
  
  function "+"(P,Q: Rational) return Rational is
  begin
    return (P.Num*Q.Den+P.Den*Q.Num)/(P.Den*Q.Den);
  end "+";
  
  function "-"(I: LInt; Q: Rational) return Rational is
  begin
    return (I*Q.Den-Q.Num)/Q.Den;
  end "-";
  
  function "-"(Q: Rational; I: LInt) return Rational is
  begin
    return (Q.Num-I*Q.Den)/Q.Den;
  end "-";
  
  function "-"(P,Q: Rational) return Rational is
  begin
    return (P.Num*Q.Den-P.Den*Q.Num)/(P.Den*Q.Den);
  end "-";
  
  function "*"(I: LInt; Q: Rational) return Rational is
  begin
    return (I*Q.Num)/Q.Den;
  end "*";
  
  function "*"(Q: Rational; I: LInt) return Rational is
  begin
    return (Q.Num*I)/Q.Den;
  end "*";
  
  function "*"(P,Q: Rational) return Rational is
  begin
    return (P.Num*Q.Num)/(P.Den*Q.Den);
  end "*";

  function "/"(I: LInt; Q: Rational) return Rational is
  begin
    return (I*Q.Den)/Q.Num;
  end "/";
  
  function "/"(Q: Rational; I: LInt) return Rational is
  begin
    return Q.Num/(Q.Den*I);
  end "/";
  
  function "/" (P,Q: Rational) return Rational is
  begin
   return (P.Num*Q.Den)/(P.Den*Q.Num);
  end "/";
  
  function "abs"(Q: Rational) return Rational is
  begin
    return (abs(Q.Num),Q.Den);
  end "abs";
  
  function "<"(P,Q: Rational) return Boolean is
    G: constant LInt := GCD(P.Den,Q.Den);
    PD: constant Lint := P.Den/G;
    QD: constant Lint := Q.Den/G;
  begin
    return (P.Num*QD)<(Q.Num*PD);
  end "<";
  
  procedure Show1(N: in String; Q: in Rational; NewLine: in Boolean := True) is
    use LInt_IO;
  begin
    Put(N);
    LInt_IO.Put(Q.Num,Width(Q.Num));
    Put("/");
    LInt_IO.Put(Q.Den,Width(Q.Den));
    if NewLine then New_Line; end if;
  end Show1;
  
end Rationals;
