with Ada.Text_IO, Globals, Flts, Flt_Vectors, Matrices, Vectors;
use Ada.Text_IO, Globals, Flts;

pragma Elaborate_All (Ada.Text_IO,Globals,Flts,Flt_Vectors,Matrices,Vectors);

generic

  type Scalar is private;
  
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is  <>;
  with function IsReal(S: Scalar) return Boolean is  <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  with package SM is new Matrices (Scalar => Scalar);
  with package SV is new Vectors (Scalar => Scalar, Component => Scalar, SM => SM);

  with function "-"(S: Scalar) return Scalar is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with function Approx(S: Scalar) return Flt is <>;

package ScalVectors is

  use SM,SV;
  
  procedure Norm1(V: in Vector; S: in out Scalar);                --- sum
  function Norm1(V: Vector) return Scalar;                        --- sum
  procedure WNorm1(W,V: in Vector; S: in out Scalar);             --- weighted sum
  procedure WNorm1(W: in RadVec; V: in Vector; S: in out Scalar); --- weighted sum
  function WNorm1(W: RadVec; V: Vector) return Scalar;            --- weighted sum
  procedure WBallAt0(W: in RadVec; R: in Flt; V: in out Vector);
  procedure Norm1(A: in Matrix; S: in out Scalar);                --- sum
  function Norm1(A: Matrix) return Scalar;                        --- sum
  procedure WNorm1(W: in Vector; A: in Matrix; S: in out Scalar); --- weighted sum
  procedure WNorm1(W: in RadVec; A: in Matrix; S: in out Scalar); --- weighted sum
  function WNorm1(W: RadVec; A: Matrix) return Scalar;            --- weighted sum
  procedure WBallAt0(W: in RadVec; R: in Flt; A: in out Matrix);
  function Norm2(V: Vector) return Scalar;
  function NormI(V: Vector) return Scalar;                        --- max
  
  subtype FltVec is Flt_Vectors.Vector;
  
  procedure Center(V: in out Vector);
  procedure Center(A: in out Matrix);
  procedure Dot(V1,V2: in Vector; S: in out Scalar);
  function Dot(V1,V2: Vector) return Scalar;         --- linear in both args
  function "*"(V1,V2: Vector) return Scalar;         --- antilinear in V1
  function "*"(S: Scalar; U: FltVec) return Vector;
  procedure Truncate(Eps: in Radius; V: in out Vector);
  procedure Truncate(Eps: in Radius; A: in out Matrix);
  procedure SetRow(I: in Integer; V: in Vector; A: in out Matrix);
  procedure SetColumn(J: in Integer; V: in Vector; A: in out Matrix);
  procedure GetRow(I: in Integer; A: in Matrix; V: in out Vector);
  function GetRow(I: Integer; A: Matrix) return Vector;
  procedure GetColumn(J: in Integer; A: in Matrix; V: in out Vector);
  function GetColumn(J: Integer; A: Matrix) return Vector;
  procedure Symmetrize(A: in out Matrix);
  function AntiSymm(A: in Matrix) return Matrix;
  procedure Orthogonalize(A: in out Matrix);                       -- Gram-Schmidt
  procedure Solve(B: in Vector; A: in Matrix; X: in out Vector);   -- Solve A*X=B
  procedure Invert(A: in out Matrix);                              -- Invert A
  
--- purely numerical  
  procedure InvNewton(A: in Matrix; Ai: in out Matrix; Iter: in Positive); -- needs approx Ai
  procedure InvNewton(A: in out Matrix; Iter: in Positive);
  procedure CheckInv(W: in Vector; A,Ai: in Matrix);                       -- show norm of A*Ai-Id
  
  type Diag is new Vector;

  function GetDiag(A: in Matrix) return Diag;
  procedure Add(D: in Diag; A: in out Matrix);                     -- A := D+A
  procedure Mult(D: in Diag; W: in out Vector);                    -- W := D*W
  procedure Mult(D1: in Diag; D2: in out Diag);                    -- D2 := D1*D2
  procedure Mult(D: in Diag; A: in out Matrix);                    -- A := D*A
  procedure Mult(A: in out Matrix; D: in Diag);                    -- A := A*D
  function "*"(D: Diag; A: Matrix) return Matrix;                  -- return D*A
  function "*"(A: Matrix; D: Diag) return Matrix;                  -- return A*D
  procedure Invert(D: in out Diag);                                -- invert D
  function Inverse(D: in Diag) return Diag;                        -- return D^{-1}

--- baby cases

  subtype Vector2 is Vector(1..2);
  subtype Vector3 is Vector(1..3);
  
  procedure Inv2(A1: in Matrix2; A2: in out Matrix2; Det: in out Scalar);
  function Inv2(A: Matrix2) return Matrix2;
  procedure Solve2(B: in Vector2; A: in Matrix2; X: in out Vector2);    --- solve A*X=B
  procedure LQuot2(A1,A2: in Matrix2; A3: in out Matrix2);              --- A3 := (1/A2)A1
  procedure EigenVals2(A: in Matrix2; U1,U2,V: in out Scalar);          --- see Roots2
--- only for real Scalar:
  procedure EigenVals2(A: in Matrix2; Wr,Wi: in out Vector2);
  procedure Eigen2(A: in Matrix2; Wr,Wi: in out Vector2; Vr,Vi: in out Matrix2);
  procedure Eigen2(A: in Matrix2; W: in out Vector2; V: in out Matrix2); -- if A is real symmetric

  procedure Inv3(A1: in Matrix3; A2: in out Matrix3; Det: in out Scalar);
  function Inv3(A: Matrix3) return Matrix3;
  procedure Solve3(B: in Vector3; A: in Matrix3; X: in out Vector3);    --- solve A*X=B
  procedure LQuot3(A1,A2: in Matrix3; A3: in out Matrix3);              --- A3 := (1/A2)A1
  procedure EigenVals3(A: in Matrix3; U0,U1,U2,V: in out Scalar);       --- see Roots3
--- only for real Scalar:
  procedure EigenVals3(A: in Matrix3; Wr,Wi: in out Vector3);
  procedure Eigen3(A: in Matrix3; W: in out Scalar; V: in out Vector3); --- largest real eigenvalue, and eigenvector
  procedure Eigen3(A: in Matrix3; Wr,Wi: in out Vector3; Vr,Vi: in out Matrix3);
  
--- for unusual cases
  procedure ImproveEigenVals2(A: in Matrix2; U1,U2: in out Scalar; V: in Scalar; Steps: in Natural := 64; Parallel: in Boolean := False);
  procedure ImproveEigenVals3(A: in Matrix3; U0,U1,U2: in out Scalar; V: in Scalar; Steps: in Natural := 64; Parallel: in Boolean := False);
  
--- just for info
  procedure ShowEigenvals3(N: in String; A: in Matrix3);
  procedure EigenQuality(Ar,Ai: in Matrix; Wr,Wi: in Scalar; Vr,Vi: in Vector); -- norm of (A-W)*V
  procedure EigenQuality(A: in Matrix; Wr,Wi: in Scalar; Vr,Vi: in Vector);
  procedure EigenQuality(Ar,Ai: in Matrix; Wr,Wi: in Vector; Vr,Vi: in Matrix);
  procedure EigenQuality(A: in Matrix; Wr,Wi: in Vector; Vr,Vi: in Matrix);
  
--- used in FindZero2 and FindZero3
  Norm_Weight: RadVec(1..3) := (One,One,One);
  
  generic
    with procedure F(V: in Vector2; W: in out Vector2);
    with procedure DF(V: in Vector2; A: in out Matrix2);
  procedure FindZero2(V: in out Vector2; Steps: in Natural := 64; Parallel: in Boolean := False; Verbosity: in Natural := 0);
  
  generic
    with procedure F(V: in Vector3; W: in out Vector3);
    with procedure DF(V: in Vector3; A: in out Matrix3);
  procedure FindZero3(V: in out Vector3; Steps: in Natural := 64; Parallel: in Boolean := False; Verbosity: in Natural := 0);
  
private

  pragma Inline (IsZero,SetZero,Scal,Neg,Add,Sub,Mult,AddProd,Adjoint);

  type Pivot is array(Integer range <>) of Integer;

  function IMin(I1,I2: Integer) return Integer renames Integer'Min;
  function IMax(I1,I2: Integer) return Integer renames Integer'Max;
  procedure CheckDim(N: in Positive; V: in Vector);
  procedure CheckDim(N: in Positive; A: in Matrix);

  procedure Dgefa(A: in out Matrix; Pvt: out Pivot);
  procedure Dgesl(Pvt: in Pivot; A: in Matrix; B: in out Vector);
  procedure Dgedi(Pvt: in Pivot; A: in out Matrix; B: in out Vector);
  procedure Solve0(B: in Vector; A: in Matrix; X: in out Vector);
  procedure Invert0(A: in out Matrix);
  
  procedure InvNewtonStep(A: in Matrix; Ai,B1,B2: in out Matrix; Err: out Radius);
  procedure RealEigenvec3(A: in Matrix3; W: in Scalar; V: in out Vector3);
  procedure ComplexEigenvec3(A: in Matrix3; Wr,Wi: in Scalar; Vr,Vi: in out Vector3);

  One_Scalar:   constant Scalar      := Scal(One);
  Half_Scalar:  constant Scalar      := Scal(Half);
  SInfo:        constant Scalar_Info := Info(One_Scalar);
  STrunc:       constant Boolean     := SInfo.IsNumeric;

  pragma Inline_Always (Swap,Neg,Add,Sub,Mult,AddProd,Scal,CheckDim);
  
--- just for 2x2 and 3x3
  procedure Add(I: in Integer; A: in out Matrix) renames AddConst;
  procedure Quot(B: in Vector; A: in Matrix; X: in out Vector);
  procedure Quot(A1,A2: in Matrix; A3: in out Matrix);
  function MaxNorm(V: Vector) return Radius;
  function MaxNorm(A: Matrix) return Radius;
  procedure BallAt0(R: in Flt; V: in out Vector);
  procedure BallAt0(R: in Flt; A: in out Matrix);
  
end ScalVectors;
