with Ada.Text_IO, VGeneric, Matrices;
use Ada.Text_IO, VGeneric;

pragma Elaborate_All (Ada.Text_IO,Matrices,VGeneric);

generic

  type Scalar is private;

  with function "="(S1,S2: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is  <>;
  with function IsReal(S: Scalar) return Boolean is  <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Assign(I: in Integer; S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  type Component is private;

  with procedure Mult(S: in Scalar; C: in out Component) is <>;
  with procedure Prod(S: in Scalar; C1: in Component; C2: in out Component) is <>;
  with procedure AddProd(S: in Scalar; C1: in Component; C2: in out Component) is <>;

--  with function "-"(S: Component) return Component is <>;
  with function "="(C1,C2: Component) return Boolean is <>;
  with function IsZero(C: Component) return Boolean is  <>;
  with function IsReal(C: Component) return Boolean is  <>;
  with procedure SetZero(C: in out Component) is <>;
  with procedure Copy(C1: in Component; C2: in out Component) is <>;
  with procedure Swap(C1,C2: in out Component) is <>;
  with procedure Neg(C: in out Component) is <>;
  with procedure Neg(C1: in Component; C2: in out Component) is <>;
  with procedure Add(C1: in Component; C2: in out Component) is <>;
  with procedure Sum(C1,C2: in Component; C3: in out Component) is <>;
  with procedure Sub(C1: in Component; C2: in out Component) is <>;
  with procedure Diff(C1,C2: in Component; C3: in out Component) is <>;
  with procedure Real_Part(C: in out Component) is <>;
  with procedure Imag_Part(C: in out Component) is <>;
  with procedure Adjoint(C: in out Component) is <>;
  with procedure Adjoint(C1: in Component; C2: in out Component) is <>;
  with procedure Show1(N: in String; C: in Component; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; C1,C2: in Component; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; C: in Component; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; C: in out Component; Decimal: in Boolean := False) is <>;

  with package SM is new Matrices (Scalar => Scalar);

package Vectors is

  use SM;

  type Vector is array(Integer range <>) of Component;
  
--- basic
  function "="(V1,V2: Vector) return Boolean;
  function IsZero is new VF_BProperty (Component,Vector,IsZero);
  function IsReal is new VF_BProperty (Component,Vector,IsReal);
  procedure SetZero is new VP_Modify1 (Component,Vector,SetZero);
  procedure SetZero(F,L: in Integer; V: in out Vector);
  function ZeroVector(F,L: Integer) return Vector;
  procedure Copy is new VP_Map1 (Component,Vector,Copy);
  procedure Swap is new VP_Modify2 (Component,Vector,Swap);
  procedure Neg is new VP_Modify1 (Component,Vector,Neg);
  procedure Neg is new VP_Map1 (Component,Vector,Neg);
  function "-" is new VF_Map1 (Component,Vector,Neg);
  procedure Add is new VP_Map1 (Component,Vector,Add);
  procedure Sum is new VP_Map2 (Component,Vector,Sum);
  function "+" is new VF_Map2 (Component,Vector,Sum);
  procedure Sub is new VP_Map1 (Component,Vector,Sub);
  procedure Diff is new VP_Map2 (Component,Vector,Diff);
  function "-" is new VF_Map2 (Component,Vector,Diff);
  procedure Mult(S: in Scalar; V: in out Vector);
  procedure Prod(S: in Scalar; V1: in Vector; V2: in out Vector);
  function "*"(S: Scalar; V: Vector) return Vector;
  procedure AddProd(S: in Scalar; V1: in Vector; V2: in out Vector);
  procedure Real_Part is new VP_Modify1 (Component,Vector,Real_Part);
  procedure Imag_Part is new VP_Modify1 (Component,Vector,Imag_Part);
  procedure Adjoint is new VP_Modify1 (Component,Vector,Adjoint);
  procedure Adjoint is new VP_Map1 (Component,Vector,Adjoint);
  
  procedure Mult(A: in Matrix; V,Tmp: in out Vector);
  procedure Mult(A: in Matrix; V: in out Vector);
  procedure Prod(A: in Matrix; V1: in Vector; V2: in out Vector);
  function "*"(A: Matrix; V: Vector) return Vector;
  procedure AddProd(A: in Matrix; V1: in Vector; V2: in out Vector);

--- i/o and conversion
  procedure Show1(N: in String; V: in Vector; Hide0: in Boolean := True);
  procedure Show2(N: in String; V1,V2: in Vector; Hide0: in Boolean := True);
  procedure PutDim(F: in File_Type; L: in Positive);
  procedure PutCoeffs(F: in File_Type; V: in Vector; Decimal: in Boolean := False);
  procedure Put(F: in File_Type; V: in Vector; Decimal: in Boolean := False);
  function Get(F: File_Type; First: Integer; Decimal: Boolean := False) return Vector;
  procedure Get(F: in File_Type; V: in out Vector; Decimal: in Boolean := False);
  procedure Write(FileName: in String; V: in Vector; Decimal: in Boolean := False);
  function Read(FileName: String; First: Integer; Decimal: Boolean := False) return Vector;
  procedure Read(FileName: in String; V: in out Vector; Decimal: in Boolean := False);
  procedure Append(FileName: in String; V: in Vector; Decimal: in Boolean := False);

--- auxiliary
  function EffFirst(V: Vector) return Integer;
  function EffLast(V: Vector) return Integer;
  procedure Trim(F,L: in Integer; V: in out Vector);
  
  subtype Polynom1 is Vector; --- Natural range
  
  function EffDeg(P: Polynom1) return Natural;
  procedure SetCoeff(I: in Natural; C: in Component; P: in out Polynom1);
  procedure Coefficient(P: in Polynom1; I: in Natural; C: in out Component);
  procedure Evaluate(P: in Polynom1; S: in Scalar; C: in out Component; Deg: in Natural); -- value of P at S
  procedure Evaluate(P: in Polynom1; S: in Scalar; C: in out Component);                  -- value of P at S
  function Evaluate(P: Polynom1; S: Scalar; Deg: Natural) return Component;               -- value of P at S
  function Evaluate(P: Polynom1; S: Scalar) return Component;                             -- value of P at S
  procedure Translate(P1: in Polynom1; S: in Scalar; P2: in out Polynom1);                -- P2(x) := P1(x+S)
  
  type Polynom2 is new Vector; --- index = J+D*(D+1)/2   where D=I+J
  
  function MaxDeg(P: Polynom2) return Natural;
  function EffDeg(P: Polynom2) return Natural;
  procedure SetCoeff(I,J: in Natural; C: in Component; P: in out Polynom2);
  procedure Coefficient(P: in Polynom2; I,J: in Natural; C: in out Component);
  procedure Evaluate(P: in Polynom2; S1,S2: in Scalar; C: in out Component);
  procedure Show1(N: in String; P: in Polynom2; Hide0: in Boolean := True);
  procedure Show2(N: in String; P1,P2: in Polynom2; Hide0: in Boolean := True);
  
  pragma Inline_Always(EffFirst,EffLast,EffDeg);
  
end Vectors;
